"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import random
from pathlib import Path
from typing import List
import numpy as np
import torch
from safetensors.torch import load_model
from torch.cuda.amp import GradScaler
from .utils import (MODEL_NAME, OPTIMIZER_NAME, RNG_STATE_NAME, SAFE_MODEL_NAME, SAFE_WEIGHTS_NAME, SAMPLER_NAME, SCALER_NAME, SCHEDULER_NAME, WEIGHTS_NAME,
get_pretty_name, is_mlu_available, is_torch_xla_available, is_xpu_available, save)
if is_torch_xla_available(): import torch_xla.core.xla_model as xm
from .logging import get_logger
from .state import PartialState
logger = get_logger(__name__)
def save_accelerator_state(output_dir: str, model_states: List[dict], optimizers: list, schedulers: list, dataloaders: list, process_index: int,
step: int, scaler: GradScaler = None, save_on_each_node: bool = False, safe_serialization: bool = True):
    output_dir = Path(output_dir)
    for i, state in enumerate(model_states):
        weights_name = WEIGHTS_NAME if not safe_serialization else SAFE_WEIGHTS_NAME
        if i > 0: weights_name = weights_name.replace(".", f"_{i}.")
        output_model_file = output_dir.joinpath(weights_name)
        save(state, output_model_file, save_on_each_node=save_on_each_node, safe_serialization=safe_serialization)
    for i, opt in enumerate(optimizers):
        state = opt.state_dict()
        optimizer_name = f"{OPTIMIZER_NAME}.bin" if i == 0 else f"{OPTIMIZER_NAME}_{i}.bin"
        output_optimizer_file = output_dir.joinpath(optimizer_name)
        save(state, output_optimizer_file, save_on_each_node=save_on_each_node, safe_serialization=False)
    for i, scheduler in enumerate(schedulers):
        state = scheduler.state_dict()
        scheduler_name = f"{SCHEDULER_NAME}.bin" if i == 0 else f"{SCHEDULER_NAME}_{i}.bin"
        output_scheduler_file = output_dir.joinpath(scheduler_name)
        save(state, output_scheduler_file, save_on_each_node=save_on_each_node, safe_serialization=False)
    for i, dataloader in enumerate(dataloaders):
        sampler_name = f"{SAMPLER_NAME}.bin" if i == 0 else f"{SAMPLER_NAME}_{i}.bin"
        output_sampler_file = output_dir.joinpath(sampler_name)
        from .data_loader import IterableDatasetShard, SeedableRandomSampler
        if isinstance(dataloader.dataset, IterableDatasetShard):
            sampler = dataloader.get_sampler()
            if isinstance(sampler, SeedableRandomSampler): save(sampler, output_sampler_file, save_on_each_node=save_on_each_node, safe_serialization=False)
        if getattr(dataloader, "use_stateful_dataloader", False):
            dataloader_state_dict_name = "dl_state_dict.bin" if i == 0 else f"dl_state_dict_{i}.bin"
            output_dataloader_state_dict_file = output_dir.joinpath(dataloader_state_dict_name)
            state_dict = dataloader.state_dict()
            torch.save(state_dict, output_dataloader_state_dict_file)
    if scaler is not None:
        state = scaler.state_dict()
        output_scaler_file = output_dir.joinpath(SCALER_NAME)
        torch.save(state, output_scaler_file)
    states = {}
    states_name = f"{RNG_STATE_NAME}_{process_index}.pkl"
    states["step"] = step
    states["random_state"] = random.getstate()
    states["numpy_random_seed"] = np.random.get_state()
    states["torch_manual_seed"] = torch.get_rng_state()
    if is_xpu_available(): states["torch_xpu_manual_seed"] = torch.xpu.get_rng_state_all()
    if is_mlu_available(): states["torch_mlu_manual_seed"] = torch.mlu.get_rng_state_all()
    else: states["torch_cuda_manual_seed"] = torch.cuda.get_rng_state_all()
    if is_torch_xla_available(): states["xm_seed"] = xm.get_rng_state()
    output_states_file = output_dir.joinpath(states_name)
    torch.save(states, output_states_file)
    return output_dir
def load_accelerator_state(input_dir, models, optimizers, schedulers, dataloaders, process_index, scaler=None, map_location=None, **load_model_func_kwargs):
    override_attributes = dict()
    if map_location not in [None, "cpu", "on_device"]: raise TypeError("Unsupported optimizer map location passed, please choose one of `None`, `'cpu'`, or `'on_device'`")
    if map_location is None: map_location = "cpu"
    elif map_location == "on_device": map_location = PartialState().device
    input_dir = Path(input_dir)
    for i, model in enumerate(models):
        ending = f"_{i}" if i > 0 else ""
        input_model_file = input_dir.joinpath(f"{SAFE_MODEL_NAME}{ending}.safetensors")
        if input_model_file.exists(): load_model(model, input_model_file, device=str(map_location), **load_model_func_kwargs)
        else:
            input_model_file = input_dir.joinpath(f"{MODEL_NAME}{ending}.bin")
            state_dict = torch.load(input_model_file, map_location=map_location)
            model.load_state_dict(state_dict, **load_model_func_kwargs)
    for i, opt in enumerate(optimizers):
        optimizer_name = f"{OPTIMIZER_NAME}.bin" if i == 0 else f"{OPTIMIZER_NAME}_{i}.bin"
        input_optimizer_file = input_dir.joinpath(optimizer_name)
        optimizer_state = torch.load(input_optimizer_file, map_location=map_location)
        optimizers[i].load_state_dict(optimizer_state)
    for i, scheduler in enumerate(schedulers):
        scheduler_name = f"{SCHEDULER_NAME}.bin" if i == 0 else f"{SCHEDULER_NAME}_{i}.bin"
        input_scheduler_file = input_dir.joinpath(scheduler_name)
        scheduler.load_state_dict(torch.load(input_scheduler_file))
    for i, dataloader in enumerate(dataloaders):
        sampler_name = f"{SAMPLER_NAME}.bin" if i == 0 else f"{SAMPLER_NAME}_{i}.bin"
        input_sampler_file = input_dir.joinpath(sampler_name)
        from .data_loader import IterableDatasetShard, SeedableRandomSampler
        if isinstance(dataloader.dataset, IterableDatasetShard):
            sampler = dataloader.get_sampler()
            if isinstance(sampler, SeedableRandomSampler): sampler = dataloader.set_sampler(torch.load(input_sampler_file))
        if getattr(dataloader, "use_stateful_dataloader", False):
            dataloader_state_dict_name = "dl_state_dict.bin" if i == 0 else f"dl_state_dict_{i}.bin"
            input_dataloader_state_dict_file = input_dir.joinpath(dataloader_state_dict_name)
            if input_dataloader_state_dict_file.exists():
                state_dict = torch.load(input_dataloader_state_dict_file)
                dataloader.load_state_dict(state_dict)
    if scaler is not None:
        input_scaler_file = input_dir.joinpath(SCALER_NAME)
        scaler.load_state_dict(torch.load(input_scaler_file))
    try:
        states = torch.load(input_dir.joinpath(f"{RNG_STATE_NAME}_{process_index}.pkl"))
        if "step" in states: override_attributes["step"] = states["step"]
        random.setstate(states["random_state"])
        np.random.set_state(states["numpy_random_seed"])
        torch.set_rng_state(states["torch_manual_seed"])
        if is_xpu_available(): torch.xpu.set_rng_state_all(states["torch_xpu_manual_seed"])
        if is_mlu_available(): torch.mlu.set_rng_state_all(states["torch_mlu_manual_seed"])
        else: torch.cuda.set_rng_state_all(states["torch_cuda_manual_seed"])
        if is_torch_xla_available(): xm.set_rng_state(states["xm_seed"])
    except Exception: pass
    return override_attributes
def save_custom_state(obj, path, index: int = 0, save_on_each_node: bool = False):
    save_location = Path(path) / f"custom_checkpoint_{index}.pkl"
    save(obj.state_dict(), save_location, save_on_each_node=save_on_each_node)
def load_custom_state(obj, path, index: int = 0):
    load_location = f"{path}/custom_checkpoint_{index}.pkl"
    obj.load_state_dict(torch.load(load_location, map_location="cpu"))
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
from sapiens_accelerator.test_utils import execute_subprocess_async, path_in_sapiens_accelerator_package
def test_command_parser(subparsers=None):
    if subparsers is not None: parser = subparsers.add_parser("test")
    else: parser = argparse.ArgumentParser("SapiensAccelerator test command")
    parser.add_argument("--config_file", default=None, help=(""))
    if subparsers is not None: parser.set_defaults(func=test_command)
    return parser
def test_command(args): pass
def main():
    parser = test_command_parser()
    args = parser.parse_args()
    test_command(args)
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import pickle
import tempfile
import warnings
from typing import List
from unittest.mock import Mock
import torch
from torch.utils.data import (BatchSampler, DataLoader, Dataset, IterableDataset, RandomSampler, TensorDataset, default_collate)
from sapiens_accelerator.accelerator import Accelerator, DataLoaderConfiguration
from sapiens_accelerator.utils.dataclasses import DistributedType
NUM_ELEMENTS = 22
NUM_WORKERS = 4
BATCH_SIZE = 4
class DummyDataset(Dataset):
    def __len__(self): return NUM_ELEMENTS
    def __getitem__(self, index):
        squeeze = False
        if isinstance(index, int):
            index = [index]
            squeeze = True
        elif isinstance(index, slice): index = list(range(*index.indices(self.size)))
        else: index = list(index)
        batch = [{"index": i, "label": i % 2, "random_augmentation": torch.rand(1).item()} for i in index]
        if squeeze: batch = batch[0]
        return batch
class DummyIterableDataset(IterableDataset):
    def __init__(self, data): self.data = data
    def __iter__(self): yield from self.data
def create_accelerator(even_batches=True):
    dataloader_config = DataLoaderConfiguration(even_batches=even_batches)
    accelerator = Accelerator(dataloader_config=dataloader_config)
    assert accelerator.num_processes == 2, "this script expects that two GPUs are available"
    return accelerator
def create_dataloader(accelerator: Accelerator, dataset_size: int, batch_size: int, iterable: bool = False, shuffle: bool = False):
    values = torch.as_tensor(range(dataset_size))
    if shuffle: values = values[torch.randperm(values.size(0))]
    if iterable: dataset = DummyIterableDataset(values)
    else: dataset = TensorDataset(torch.as_tensor(range(dataset_size)))
    dl = DataLoader(dataset, batch_size=batch_size)
    dl = accelerator.prepare(dl)
    return dl
def verify_dataloader_batch_sizes(accelerator: Accelerator, dataset_size: int, batch_size: int, process_0_expected_batch_sizes: List[int], process_1_expected_batch_sizes: List[int]):
    dl = create_dataloader(accelerator=accelerator, dataset_size=dataset_size, batch_size=batch_size)
    batch_sizes = [len(batch[0]) for batch in dl]
    if accelerator.process_index == 0: assert batch_sizes == process_0_expected_batch_sizes
    elif accelerator.process_index == 1: assert batch_sizes == process_1_expected_batch_sizes
def test_default_ensures_even_batch_sizes():
    accelerator = create_accelerator()
    verify_dataloader_batch_sizes(accelerator, dataset_size=3, batch_size=1, process_0_expected_batch_sizes=[1, 1], process_1_expected_batch_sizes=[1, 1])
    verify_dataloader_batch_sizes(accelerator, dataset_size=7, batch_size=2, process_0_expected_batch_sizes=[2, 2], process_1_expected_batch_sizes=[2, 2])
def test_can_disable_even_batches():
    accelerator = create_accelerator(even_batches=False)
    verify_dataloader_batch_sizes(accelerator, dataset_size=3, batch_size=1, process_0_expected_batch_sizes=[1, 1], process_1_expected_batch_sizes=[1])
    verify_dataloader_batch_sizes(accelerator, dataset_size=7, batch_size=2, process_0_expected_batch_sizes=[2, 2], process_1_expected_batch_sizes=[2, 1])
def test_can_join_uneven_inputs():
    accelerator = create_accelerator(even_batches=False)
    model = torch.nn.Linear(1, 1)
    ddp_model = accelerator.prepare(model)
    dl = create_dataloader(accelerator, dataset_size=3, batch_size=1)
    batch_idxs = []
    with accelerator.join_uneven_inputs([ddp_model]):
        for batch_idx, batch in enumerate(dl):
            output = ddp_model(batch[0].float())
            loss = output.sum()
            loss.backward()
            batch_idxs.append(batch_idx)
    accelerator.wait_for_everyone()
    if accelerator.process_index == 0: assert batch_idxs == [0, 1]
    elif accelerator.process_index == 1: assert batch_idxs == [0]
def test_join_raises_warning_for_non_ddp_distributed(accelerator):
    with warnings.catch_warnings(record=True) as w:
        with accelerator.join_uneven_inputs([Mock()]): pass
        assert issubclass(w[-1].category, UserWarning)
        assert "only supported for multi-GPU" in str(w[-1].message)
def test_join_can_override_even_batches():
    default_even_batches = True
    overridden_even_batches = False
    accelerator = create_accelerator(even_batches=default_even_batches)
    model = torch.nn.Linear(1, 1)
    ddp_model = accelerator.prepare(model)
    train_dl = create_dataloader(accelerator, dataset_size=3, batch_size=1)
    valid_dl = create_dataloader(accelerator, dataset_size=3, batch_size=1)
    with accelerator.join_uneven_inputs([ddp_model], even_batches=overridden_even_batches):
        train_dl_overridden_value = train_dl.batch_sampler.even_batches
        valid_dl_overridden_value = valid_dl.batch_sampler.even_batches
    assert train_dl_overridden_value == overridden_even_batches
    assert valid_dl_overridden_value == overridden_even_batches
    assert train_dl.batch_sampler.even_batches == default_even_batches
    assert valid_dl.batch_sampler.even_batches == default_even_batches
def test_join_can_override_for_mixed_type_dataloaders():
    default_even_batches = True
    overridden_even_batches = False
    accelerator = create_accelerator(even_batches=default_even_batches)
    model = torch.nn.Linear(1, 1)
    ddp_model = accelerator.prepare(model)
    create_dataloader(accelerator, dataset_size=3, batch_size=1, iterable=True)
    batch_dl = create_dataloader(accelerator, dataset_size=3, batch_size=1)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        try:
            with accelerator.join_uneven_inputs([ddp_model], even_batches=overridden_even_batches): batch_dl_overridden_value = batch_dl.batch_sampler.even_batches
        except AttributeError: raise AssertionError
    assert batch_dl_overridden_value == overridden_even_batches
    assert batch_dl.batch_sampler.even_batches == default_even_batches
def test_join_raises_warning_for_iterable_when_overriding_even_batches():
    accelerator = create_accelerator()
    model = torch.nn.Linear(1, 1)
    ddp_model = accelerator.prepare(model)
    create_dataloader(accelerator, dataset_size=3, batch_size=1, iterable=True)
    with warnings.catch_warnings(record=True) as w:
        with accelerator.join_uneven_inputs([ddp_model], even_batches=False): pass
        assert issubclass(w[-1].category, UserWarning)
        assert "only supported for map-style datasets" in str(w[-1].message)
def test_pickle_accelerator():
    accelerator = create_accelerator()
    data_loader = create_dataloader(accelerator, dataset_size=32, batch_size=4)
    _ = accelerator.prepare(data_loader)
    pickled_accelerator = pickle.dumps(accelerator)
    unpickled_accelerator = pickle.loads(pickled_accelerator)
    assert accelerator.state.__dict__ == unpickled_accelerator.state.__dict__
def test_data_loader(data_loader, accelerator):
    data_loader = accelerator.prepare(data_loader)
    all_examples = []
    for i, batch in enumerate(data_loader):
        index, _ = accelerator.gather_for_metrics((batch["index"], batch["label"]))
        all_examples.extend(index.detach().cpu().numpy().tolist())
    sorted_all_examples = sorted(all_examples)
    assert (len(set(sorted_all_examples)) == NUM_ELEMENTS), "Not all the dataset elements have been iterated in an epoch due to duplication of samples across processes."
def test_stateful_dataloader(accelerator):
    old_dataloader_config = accelerator.dataloader_config
    try:
        accelerator.dataloader_config = DataLoaderConfiguration(use_stateful_dataloader=True)
        prepared_dl = create_dataloader(accelerator, dataset_size=32 * accelerator.num_processes, batch_size=4, iterable=True, shuffle=True)
        untrained_batches = []
        total_batches = 32 * accelerator.num_processes // (4 * accelerator.num_processes)
        last_batch_num = total_batches - 1
        for step, batch in enumerate(prepared_dl):
            if step == last_batch_num - 1: state_dict = prepared_dl.state_dict()
            if step >= last_batch_num: untrained_batches.append(batch)
        not_skipped_batches = accelerator.gather(untrained_batches)
        prepared_dl.load_state_dict(state_dict)
        resumed_batches = []
        for batch in prepared_dl: resumed_batches.append(batch)
        resumed_batches = accelerator.gather(resumed_batches)
        for b1, b2 in zip(not_skipped_batches, resumed_batches):
            for v1, v2 in zip(b1, b2): assert torch.equal(v1, v2), f"Batch {b1} and {b2} are not equal"
    finally: accelerator.dataloader_config = old_dataloader_config
def test_stateful_dataloader_save_state(accelerator):
    old_dataloader_config = accelerator.dataloader_config
    try:
        with tempfile.TemporaryDirectory() as tmpdir:
            accelerator.dataloader_config = DataLoaderConfiguration(use_stateful_dataloader=True)
            prepared_dl = create_dataloader(accelerator, dataset_size=32 * accelerator.num_processes, batch_size=4, iterable=True, shuffle=True)
            untrained_batches = []
            total_batches = 32 * accelerator.num_processes // (4 * accelerator.num_processes)
            last_batch_num = total_batches - 1
            for step, batch in enumerate(prepared_dl):
                if step == last_batch_num - 1: accelerator.save_state(tmpdir)
                if step >= last_batch_num: untrained_batches.append(batch)
            not_skipped_batches = accelerator.gather(untrained_batches)
            accelerator.load_state(tmpdir)
            resumed_batches = []
            for batch in prepared_dl: resumed_batches.append(batch)
            resumed_batches = accelerator.gather(resumed_batches)
            for b1, b2 in zip(not_skipped_batches, resumed_batches):
                for v1, v2 in zip(b1, b2): assert torch.equal(v1, v2), f"Batch {b1} and {b2} are not equal"
    finally: accelerator.dataloader_config = old_dataloader_config
def main():
    accelerator = create_accelerator()
    torch.manual_seed(accelerator.process_index)
    test_default_ensures_even_batch_sizes()
    test_can_disable_even_batches()
    test_can_join_uneven_inputs()
    test_join_can_override_even_batches()
    test_join_can_override_for_mixed_type_dataloaders()
    test_join_raises_warning_for_iterable_when_overriding_even_batches()
    original_state = accelerator.state.distributed_type
    accelerator.state.distributed_type = DistributedType.FSDP
    test_join_raises_warning_for_non_ddp_distributed(accelerator)
    accelerator.state.distributed_type = original_state
    accelerator.print("Test pickling an accelerator")
    test_pickle_accelerator()
    dataset = DummyDataset()
    loader = DataLoader(dataset, shuffle=False, batch_size=BATCH_SIZE, num_workers=NUM_WORKERS)
    test_data_loader(loader, accelerator)
    loader = DataLoader(dataset, shuffle=True, batch_size=BATCH_SIZE, num_workers=NUM_WORKERS)
    test_data_loader(loader, accelerator)
    sampler = BatchSampler(RandomSampler(dataset), batch_size=BATCH_SIZE, drop_last=False)
    loader = DataLoader(dataset, batch_sampler=sampler, num_workers=NUM_WORKERS)
    test_data_loader(loader, accelerator)
    sampler = BatchSampler(RandomSampler(dataset), batch_size=BATCH_SIZE, drop_last=False)
    loader = DataLoader(dataset, sampler=sampler, batch_size=None, collate_fn=default_collate, num_workers=NUM_WORKERS)
    test_data_loader(loader, accelerator)
    test_stateful_dataloader(accelerator)
    test_stateful_dataloader_save_state(accelerator)
    accelerator.end_training()
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
from .config import config_command_parser
from .config_args import default_config_file, load_config_from_file
from .default import default_command_parser
from .update import update_command_parser
def get_config_parser(subparsers=None):
    parent_parser = argparse.ArgumentParser(add_help=False, allow_abbrev=False)
    config_parser = config_command_parser(subparsers)
    subcommands = config_parser.add_subparsers(title="subcommands", dest="subcommand")
    default_command_parser(subcommands, parents=[parent_parser])
    update_command_parser(subcommands, parents=[parent_parser])
    return config_parser
def main():
    config_parser = get_config_parser()
    args = config_parser.parse_args()
    if not hasattr(args, "func"):
        config_parser.print_help()
        exit(1)
    args.func(args)
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import os
import string
import sys
ARROW_KEY_FLAG = 1 << 8
KEYMAP = {"tab": ord("\t"), "newline": ord("\r"), "esc": 27, "up": 65 + ARROW_KEY_FLAG, "down": 66 + ARROW_KEY_FLAG, "right": 67 + ARROW_KEY_FLAG,
"left": 68 + ARROW_KEY_FLAG, "mod_int": 91, "undefined": sys.maxsize, "interrupt": 3, "insert": 50, "delete": 51, "pg_up": 53, "pg_down": 54}
KEYMAP["arrow_begin"] = KEYMAP["up"]
KEYMAP["arrow_end"] = KEYMAP["left"]
if sys.platform == "win32":
    WIN_CH_BUFFER = []
    WIN_KEYMAP = {b"\xe0H": KEYMAP["up"] - ARROW_KEY_FLAG, b"\x00H": KEYMAP["up"] - ARROW_KEY_FLAG, b"\xe0P": KEYMAP["down"] - ARROW_KEY_FLAG,
    b"\x00P": KEYMAP["down"] - ARROW_KEY_FLAG, b"\xe0M": KEYMAP["right"] - ARROW_KEY_FLAG, b"\x00M": KEYMAP["right"] - ARROW_KEY_FLAG,
    b"\xe0K": KEYMAP["left"] - ARROW_KEY_FLAG, b"\x00K": KEYMAP["left"] - ARROW_KEY_FLAG}
for i in range(10): KEYMAP[str(i)] = ord(str(i))
def get_raw_chars():
    "Gets raw characters from inputs"
    if os.name == "nt":
        import msvcrt
        encoding = "mbcs"
        while msvcrt.kbhit(): msvcrt.getch()
        if len(WIN_CH_BUFFER) == 0:
            ch = msvcrt.getch()
            if ch in (b"\x00", b"\xe0"):
                ch2 = ch + msvcrt.getch()
                try:
                    chx = chr(WIN_KEYMAP[ch2])
                    WIN_CH_BUFFER.append(chr(KEYMAP["mod_int"]))
                    WIN_CH_BUFFER.append(chx)
                    if ord(chx) in (KEYMAP["insert"] - 1 << 9, KEYMAP["delete"] - 1 << 9, KEYMAP["pg_up"] - 1 << 9, KEYMAP["pg_down"] - 1 << 9): WIN_CH_BUFFER.append(chr(126))
                    ch = chr(KEYMAP["esc"])
                except KeyError: ch = ch2[1]
            else: ch = ch.decode(encoding)
        else: ch = WIN_CH_BUFFER.pop(0)
    elif os.name == "posix":
        import termios
        import tty
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            ch = sys.stdin.read(1)
        finally: termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch
def get_character():
    "Gets a character from the keyboard and returns the key code"
    char = get_raw_chars()
    if ord(char) in [KEYMAP["interrupt"], KEYMAP["newline"]]: return char
    elif ord(char) == KEYMAP["esc"]:
        combo = get_raw_chars()
        if ord(combo) == KEYMAP["mod_int"]:
            key = get_raw_chars()
            if ord(key) >= KEYMAP["arrow_begin"] - ARROW_KEY_FLAG and ord(key) <= KEYMAP["arrow_end"] - ARROW_KEY_FLAG: return chr(ord(key) + ARROW_KEY_FLAG)
            else: return KEYMAP["undefined"]
        else: return get_raw_chars()
    else:
        if char in string.printable: return char
        else: return KEYMAP["undefined"]
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import os
from typing import List
def get_function_contents_by_name(lines: List[str], name: str):
    if name != "training_function" and name != "main": raise ValueError(f"Incorrect function name passed: {name}, choose either 'main' or 'training_function'")
    good_lines, found_start = [], False
    for line in lines:
        if not found_start and f"def {name}" in line:
            found_start = True
            good_lines.append(line)
            continue
        if found_start:
            if name == "training_function" and "def main" in line: return good_lines
            if name == "main" and "if __name__" in line: return good_lines
            good_lines.append(line)
def clean_lines(lines: List[str]): return [line for line in lines if not line.lstrip().startswith("#") and line != "\n"]
def compare_against_test(base_filename: str, feature_filename: str, parser_only: bool, secondary_filename: str = None):
    with open(base_filename) as f: base_file_contents = f.readlines()
    with open(os.path.abspath(os.path.join("examples", "nlp_example.py"))) as f: full_file_contents = f.readlines()
    with open(feature_filename) as f: feature_file_contents = f.readlines()
    if secondary_filename is not None:
        with open(secondary_filename) as f: secondary_file_contents = f.readlines()
    if parser_only:
        base_file_func = clean_lines(get_function_contents_by_name(base_file_contents, "main"))
        full_file_func = clean_lines(get_function_contents_by_name(full_file_contents, "main"))
        feature_file_func = clean_lines(get_function_contents_by_name(feature_file_contents, "main"))
        if secondary_filename is not None: secondary_file_func = clean_lines(get_function_contents_by_name(secondary_file_contents, "main"))
    else:
        base_file_func = clean_lines(get_function_contents_by_name(base_file_contents, "training_function"))
        full_file_func = clean_lines(get_function_contents_by_name(full_file_contents, "training_function"))
        feature_file_func = clean_lines(get_function_contents_by_name(feature_file_contents, "training_function"))
        if secondary_filename is not None: secondary_file_func = clean_lines(get_function_contents_by_name(secondary_file_contents, "training_function"))
    _dl_line = "train_dataloader, eval_dataloader = get_dataloaders(accelerator, batch_size)\n"
    new_feature_code = []
    passed_idxs = []
    it = iter(feature_file_func)
    for i in range(len(feature_file_func) - 1):
        if i not in passed_idxs:
            line = next(it)
            if (line not in full_file_func) and (line.lstrip() != _dl_line):
                if "TESTING_MOCKED_DATALOADERS" not in line:
                    new_feature_code.append(line)
                    passed_idxs.append(i)
                else: _ = next(it)
    new_full_example_parts = []
    passed_idxs = []
    for i, line in enumerate(base_file_func):
        if i not in passed_idxs:
            if (line not in full_file_func) and (line.lstrip() != _dl_line):
                if "TESTING_MOCKED_DATALOADERS" not in line:
                    new_full_example_parts.append(line)
                    passed_idxs.append(i)
    diff_from_example = [line for line in new_feature_code if line not in new_full_example_parts]
    if secondary_filename is not None:
        diff_from_two = [line for line in full_file_contents if line not in secondary_file_func]
        diff_from_example = [line for line in diff_from_example if line not in diff_from_two]
    return diff_from_example
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import torch
from transformers import (BertConfig, BertForMaskedLM, GPT2Config, GPT2ForSequenceClassification)
from sapiens_accelerator import PartialState
from sapiens_accelerator.inference import prepare_pippy
from sapiens_accelerator.utils import DistributedType, set_seed
model_to_config = {"bert": (BertForMaskedLM, BertConfig, 512), "gpt2": (GPT2ForSequenceClassification, GPT2Config, 1024)}
def get_model_and_data_for_text(model_name, device, num_processes: int = 2):
    initializer, config, seq_len = model_to_config[model_name]
    config_args = {}
    model_config = config(**config_args)
    model = initializer(model_config)
    kwargs = dict(low=0, high=model_config.vocab_size, device=device, dtype=torch.int64, requires_grad=False)
    trace_input = torch.randint(size=(1, seq_len), **kwargs)
    inference_inputs = torch.randint(size=(num_processes, seq_len), **kwargs)
    return model, trace_input, inference_inputs
def test_bert(batch_size: int = 2):
    set_seed(42)
    state = PartialState()
    model, trace_input, inference_inputs = get_model_and_data_for_text("bert", "cpu", batch_size)
    model = prepare_pippy(model, example_args=(trace_input,), no_split_module_classes=model._no_split_modules)
    inputs = inference_inputs.to("cuda")
    with torch.no_grad(): output = model(inputs)
    if not state.is_last_process: assert output is None, "Output was not generated on just the last process!"
    else: assert output is not None, "Output was not generated in the last process!"
def test_gpt2(batch_size: int = 2):
    set_seed(42)
    state = PartialState()
    model, trace_input, inference_inputs = get_model_and_data_for_text("gpt2", "cpu", batch_size)
    model = prepare_pippy(model, example_args=(trace_input,), no_split_module_classes=model._no_split_modules)
    inputs = inference_inputs.to("cuda")
    with torch.no_grad(): output = model(inputs)
    if not state.is_last_process: assert output is None, "Output was not generated on just the last process!"
    else: assert output is not None, "Output was not generated in the last process!"
if __name__ == "__main__":
    state = PartialState()
    state.print("Testing pippy integration...")
    try:
        if state.distributed_type == DistributedType.MULTI_GPU:
            test_gpt2()
            test_bert()
    finally: state.destroy_process_group()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import torch
from sapiens_accelerator import Accelerator, DDPCommunicationHookType, DistributedDataParallelKwargs, PartialState
class MockModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        torch.manual_seed(0)
        self.p = torch.nn.Parameter(torch.randn(40, 20))
    def forward(self, x, rank): return self.p * (x ** (1 + rank))
def _run_and_get_grads(model, rank):
    torch.manual_seed(2024)
    input = torch.randn(40, 20)
    output = model(input, rank)
    output.mean().backward()
    param = next(model.parameters())
    return param.grad
def test_ddp_comm_hook(comm_hook, comm_wrapper, comm_state_option):
    ddp_kwargs = DistributedDataParallelKwargs(comm_hook=comm_hook, comm_wrapper=comm_wrapper, comm_state_option=comm_state_option)
    accelerator = Accelerator(kwargs_handlers=[ddp_kwargs])
    model = accelerator.prepare(MockModel())
    hook_grads = _run_and_get_grads(model, accelerator.local_process_index)
    reference_model = torch.nn.parallel.DistributedDataParallel(MockModel().to(accelerator.device), device_ids=[accelerator.local_process_index], output_device=accelerator.local_process_index)
    reference_grads = _run_and_get_grads(reference_model, accelerator.local_process_index)
    torch.testing.assert_close(hook_grads, reference_grads, rtol=1e-2, atol=1e-2)
def main():
    for comm_hook, comm_wrapper, comm_state_option in [(DDPCommunicationHookType.NO, DDPCommunicationHookType.NO, {}), (DDPCommunicationHookType.FP16, DDPCommunicationHookType.NO, {}),
    (DDPCommunicationHookType.BF16, DDPCommunicationHookType.NO, {}), (DDPCommunicationHookType.POWER_SGD, DDPCommunicationHookType.NO, {}),
    (DDPCommunicationHookType.POWER_SGD, DDPCommunicationHookType.FP16, {}), (DDPCommunicationHookType.POWER_SGD, DDPCommunicationHookType.BF16, {}),
    (DDPCommunicationHookType.POWER_SGD, DDPCommunicationHookType.NO, {"matrix_approximation_rank": 2}), (DDPCommunicationHookType.BATCHED_POWER_SGD, DDPCommunicationHookType.NO, {}),
    (DDPCommunicationHookType.BATCHED_POWER_SGD, DDPCommunicationHookType.FP16, {}), (DDPCommunicationHookType.BATCHED_POWER_SGD, DDPCommunicationHookType.BF16, {})]:
        test_ddp_comm_hook(comm_hook, comm_wrapper, comm_state_option)
    PartialState().destroy_process_group()
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
from ...utils.dataclasses import (ComputeEnvironment, DistributedType, DynamoBackend, FP8BackendType, PrecisionType, SageMakerDistributedType)
from ..menu import BulletMenu
DYNAMO_BACKENDS = ["EAGER", "AOT_EAGER", "INDUCTOR", "AOT_TS_NVFUSER", "NVPRIMS_NVFUSER", "CUDAGRAPHS", "OFI", "FX2TRT", "ONNXRT", "TENSORRT",
"AOT_TORCHXLA_TRACE_ONCE", "TORHCHXLA_TRACE_ONCE", "IPEX", "TVM"]
def _ask_field(input_text, convert_value=None, default=None, error_message=None):
    ask_again = True
    while ask_again:
        result = input(input_text)
        try:
            if default is not None and len(result) == 0: return default
            return convert_value(result) if convert_value is not None else result
        except Exception: pass
def _ask_options(input_text, options=[], convert_value=None, default=0):
    menu = BulletMenu(input_text, options)
    result = menu.run(default_choice=default)
    return convert_value(result) if convert_value is not None else result
def _convert_compute_environment(value):
    value = int(value)
    return ComputeEnvironment(["LOCAL_MACHINE", "AMAZON_SAGEMAKER"][value])
def _convert_distributed_mode(value):
    value = int(value)
    return DistributedType(["NO", "MULTI_CPU", "MULTI_XPU", "MULTI_GPU", "MULTI_NPU", "MULTI_MLU", "MULTI_MUSA", "XLA"][value])
def _convert_dynamo_backend(value):
    value = int(value)
    return DynamoBackend(DYNAMO_BACKENDS[value]).value
def _convert_mixed_precision(value):
    value = int(value)
    return PrecisionType(["no", "fp16", "bf16", "fp8"][value])
def _convert_sagemaker_distributed_mode(value):
    value = int(value)
    return SageMakerDistributedType(["NO", "DATA_PARALLEL", "MODEL_PARALLEL"][value])
def _convert_fp8_backend(value):
    value = int(value)
    return FP8BackendType(["TE", "MSAMP"][value])
def _convert_yes_no_to_bool(value): return {"yes": True, "no": False}[value.lower()]
class SubcommandHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def _format_usage(self, usage, actions, groups, prefix):
        usage = super()._format_usage(usage, actions, groups, prefix)
        usage = usage.replace("<command> [<args>] ", "")
        return usage
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
import json
import os
import evaluate
import torch
from datasets import load_dataset
from torch.optim import AdamW
from torch.utils.data import DataLoader
from transformers import AutoModelForSequenceClassification, AutoTokenizer, get_linear_schedule_with_warmup, set_seed
from sapiens_accelerator import Accelerator, DistributedType
from sapiens_accelerator.utils.deepspeed import DummyOptim, DummyScheduler
MAX_GPU_BATCH_SIZE = 16
EVAL_BATCH_SIZE = 32
def get_dataloaders(accelerator: Accelerator, batch_size: int = 16, model_name: str = "bert-base-cased"):
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    datasets = load_dataset("glue", "mrpc")
    def tokenize_function(examples):
        outputs = tokenizer(examples["sentence1"], examples["sentence2"], truncation=True, max_length=None)
        return outputs
    tokenized_datasets = datasets.map(tokenize_function, batched=True, remove_columns=["idx", "sentence1", "sentence2"], load_from_cache_file=False)
    tokenized_datasets = tokenized_datasets.rename_column("label", "labels")
    def collate_fn(examples):
        if accelerator.distributed_type == DistributedType.XLA: return tokenizer.pad(examples, padding="max_length", max_length=128, return_tensors="pt")
        return tokenizer.pad(examples, padding="longest", return_tensors="pt")
    train_dataloader = DataLoader(tokenized_datasets["train"], shuffle=True, collate_fn=collate_fn, batch_size=batch_size)
    eval_dataloader = DataLoader(tokenized_datasets["validation"], shuffle=False, collate_fn=collate_fn, batch_size=EVAL_BATCH_SIZE)
    return train_dataloader, eval_dataloader
def training_function(config, args):
    accelerator = Accelerator()
    lr = config["lr"]
    num_epochs = int(config["num_epochs"])
    seed = int(config["seed"])
    batch_size = int(config["batch_size"])
    model_name = args.model_name_or_path
    set_seed(seed)
    train_dataloader, eval_dataloader = get_dataloaders(accelerator, batch_size, model_name)
    model = AutoModelForSequenceClassification.from_pretrained(model_name, return_dict=True)
    optimizer_cls = (AdamW if accelerator.state.deepspeed_plugin is None or "optimizer" not in accelerator.state.deepspeed_plugin.deepspeed_config else DummyOptim)
    optimizer = optimizer_cls(params=model.parameters(), lr=lr)
    max_training_steps = len(train_dataloader) * num_epochs
    linear_decay_scheduler = False
    if (accelerator.state.deepspeed_plugin is None or "scheduler" not in accelerator.state.deepspeed_plugin.deepspeed_config):
        lr_scheduler = get_linear_schedule_with_warmup(optimizer=optimizer, num_warmup_steps=0, num_training_steps=max_training_steps)
        linear_decay_scheduler = True
    else: lr_scheduler = DummyScheduler(optimizer, total_num_steps=max_training_steps, warmup_num_steps=0)
    model, optimizer, train_dataloader, eval_dataloader, lr_scheduler = accelerator.prepare(model, optimizer, train_dataloader, eval_dataloader, lr_scheduler)
    starting_epoch = 0
    metric = evaluate.load("glue", "mrpc")
    best_performance = 0
    performance_metric = {}
    expected_lr_after_first_optim_step = lr * (1 - 1 / (max_training_steps / accelerator.num_processes / accelerator.gradient_accumulation_steps))
    lr_scheduler_check_completed = False
    for epoch in range(starting_epoch, num_epochs):
        model.train()
        for step, batch in enumerate(train_dataloader):
            with accelerator.accumulate(model):
                outputs = model(**batch)
                loss = outputs.loss
                accelerator.backward(loss)
                optimizer.step()
                lr_scheduler.step()
                optimizer.zero_grad()
                if (accelerator.sync_gradients and not lr_scheduler_check_completed and linear_decay_scheduler and accelerator.state.mixed_precision == "no"):
                    assert (lr_scheduler.get_last_lr()[0] == expected_lr_after_first_optim_step), f"Wrong lr found at second step, expected {expected_lr_after_first_optim_step}, got {lr_scheduler.get_last_lr()[0]}"
                    lr_scheduler_check_completed = True
        model.eval()
        samples_seen = 0
        for step, batch in enumerate(eval_dataloader):
            batch.to(accelerator.device)
            with torch.no_grad(): outputs = model(**batch)
            predictions = outputs.logits.argmax(dim=-1)
            predictions, references = accelerator.gather((predictions, batch["labels"]))
            if accelerator.use_distributed:
                if step == len(eval_dataloader) - 1:
                    predictions = predictions[: len(eval_dataloader.dataset) - samples_seen]
                    references = references[: len(eval_dataloader.dataset) - samples_seen]
                else: samples_seen += references.shape[0]
            metric.add_batch(predictions=predictions, references=references)
        eval_metric = metric.compute()
        accelerator.print(f"epoch {epoch}:", eval_metric)
        performance_metric[f"epoch-{epoch}"] = eval_metric["accuracy"]
        if best_performance < eval_metric["accuracy"]: best_performance = eval_metric["accuracy"]
    if linear_decay_scheduler and accelerator.state.mixed_precision == "no": assert (lr_scheduler.get_last_lr()[0] == 0), f"Wrong lr found at last step, expected 0, got {lr_scheduler.get_last_lr()[0]}"
    if args.performance_lower_bound is not None: assert (args.performance_lower_bound <= best_performance), f"Best performance metric {best_performance} is lower than the lower bound {args.performance_lower_bound}"
    accelerator.wait_for_everyone()
    if accelerator.is_main_process:
        with open(os.path.join(args.output_dir, "all_results.json"), "w") as f: json.dump(performance_metric, f)
    accelerator.end_training()
def main():
    parser = argparse.ArgumentParser(description="Simple example of training script tracking peak GPU memory usage.")
    parser.add_argument("--model_name_or_path", type=str, default="bert-base-cased", help="Path to pretrained model or model identifier from huggingface.co/models.", required=False)
    parser.add_argument("--output_dir", type=str, default=".", help="Optional save directory where all checkpoint folders will be stored. Default is the current working directory.")
    parser.add_argument("--performance_lower_bound", type=float, default=None, help="Optional lower bound for the performance metric. If set, the training will throw error when the performance metric drops below this value.")
    parser.add_argument("--num_epochs", type=int, default=3, help="Number of train epochs.")
    args = parser.parse_args()
    config = {"lr": 2e-5, "num_epochs": args.num_epochs, "seed": 42, "batch_size": 16}
    training_function(config, args)
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import importlib.metadata
import subprocess
import sys
def install_xla(upgrade: bool = False):
    in_colab = False
    if "IPython" in sys.modules: in_colab = "google.colab" in str(sys.modules["IPython"].get_ipython())
    if in_colab:
        if upgrade:
            torch_install_cmd = ["pip", "install", "-U", "torch"]
            subprocess.run(torch_install_cmd, check=True)
        torch_version = importlib.metadata.version("torch")
        torch_version_trunc = torch_version[: torch_version.rindex(".")]
        xla_wheel = f"https://storage.googleapis.com/tpu-pytorch/wheels/colab/torch_xla-{torch_version_trunc}-cp37-cp37m-linux_x86_64.whl"
        xla_install_cmd = ["pip", "install", xla_wheel]
        subprocess.run(xla_install_cmd, check=True)
    else: raise RuntimeError("`install_xla` utility works only on google colab.")
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
import os
import subprocess
import sys
from ast import literal_eval
from shutil import which
from typing import Any, Dict, List, Tuple
import torch
from ..commands.config.config_args import SageMakerConfig
from ..utils import (DynamoBackend, PrecisionType, is_fp8_available, is_ipex_available, is_mlu_available, is_musa_available, is_npu_available, is_torch_xla_available, is_xpu_available)
from ..utils.constants import DEEPSPEED_MULTINODE_LAUNCHERS
from ..utils.other import is_port_in_use, merge_dicts
from .dataclasses import DistributedType, SageMakerDistributedType
def _filter_args(args, parser, default_args=[]):
    new_args, _ = parser.parse_known_args(default_args)
    for key, value in vars(args).items():
        if key in vars(new_args).keys(): setattr(new_args, key, value)
    return new_args
def _get_mpirun_args():
    mpi_apps = [x for x in ["mpirun", "mpiexec"] if which(x)]
    if len(mpi_apps) == 0: raise OSError("mpirun or mpiexec were not found. Ensure that Intel MPI, Open MPI, or MVAPICH are installed.")
    mpi_app = mpi_apps[0]
    mpirun_version = subprocess.check_output([mpi_app, "--version"])
    if b"Open MPI" in mpirun_version: return mpi_app, "--hostfile", "-n", "--npernode", "--bind-to"
    else: return mpi_app, "-f", "-n", "-ppn", ""
def setup_fp8_env(args: argparse.Namespace, current_env: Dict[str, str]):
    prefix = "SAPIENS_ACCELERATOR_"
    for arg in vars(args):
        if arg.startswith("fp8_"):
            value = getattr(args, arg)
            if value is not None: current_env[f"{prefix}{arg.upper()}"] = str(getattr(args, arg))
    return current_env
def prepare_simple_launcher_cmd_env(args: argparse.Namespace) -> Tuple[List[str], Dict[str, str]]:
    cmd = []
    if args.no_python and args.module: raise ValueError("--module and --no_python cannot be used together")
    if args.mpirun_hostfile is not None:
        mpi_app_name, hostfile_arg, num_proc_arg, proc_per_node_arg, bind_to_arg = _get_mpirun_args()
        mpirun_ccl = getattr(args, "mpirun_ccl", None)
        bind_to = getattr(args, "bind-to", "socket")
        num_machines = args.num_machines
        num_processes = getattr(args, "num_processes", None)
        nproc_per_node = str(num_processes // num_machines) if num_processes and num_machines else "1"
        cmd += [mpi_app_name, hostfile_arg, args.mpirun_hostfile, proc_per_node_arg, nproc_per_node]
        if num_processes: cmd += [num_proc_arg, str(num_processes)]
        if bind_to_arg: cmd += [bind_to_arg, bind_to]
    if not args.no_python:
        cmd.append(sys.executable)
        if args.module: cmd.append("-m")
    cmd.append(args.training_script)
    cmd.extend(args.training_script_args)
    current_env = os.environ.copy()
    current_env["SAPIENS_ACCELERATOR_USE_CPU"] = str(args.cpu or args.use_cpu)
    if args.debug: current_env["SAPIENS_ACCELERATOR_DEBUG_MODE"] = "true"
    if args.gpu_ids != "all" and args.gpu_ids is not None:
        if is_xpu_available(): current_env["ZE_AFFINITY_MASK"] = args.gpu_ids
        elif is_mlu_available(): current_env["MLU_VISIBLE_DEVICES"] = args.gpu_ids
        elif is_musa_available(): current_env["MUSA_VISIBLE_DEVICES"] = args.gpu_ids
        elif is_npu_available(): current_env["ASCEND_RT_VISIBLE_DEVICES"] = args.gpu_ids
        else: current_env["CUDA_VISIBLE_DEVICES"] = args.gpu_ids
    if args.num_machines > 1:
        current_env["MASTER_ADDR"] = args.main_process_ip
        current_env["MASTER_PORT"] = str(args.main_process_port)
        if args.mpirun_hostfile is not None: current_env["CCL_WORKER_COUNT"] = str(mpirun_ccl)
    elif args.num_processes > 1:
        current_env["MASTER_ADDR"] = args.main_process_ip if args.main_process_ip is not None else "127.0.0.1"
        current_env["MASTER_PORT"] = str(args.main_process_port) if args.main_process_port is not None else "29500"
    try: mixed_precision = PrecisionType(args.mixed_precision.lower())
    except ValueError: raise ValueError(f"Unknown mixed_precision mode: {args.mixed_precision.lower()}. Choose between {PrecisionType.list()}.")
    current_env["SAPIENS_ACCELERATOR_MIXED_PRECISION"] = str(mixed_precision)
    if args.mixed_precision.lower() == "fp8":
        if not is_fp8_available(): raise RuntimeError("FP8 is not available on this machine. Please ensure that either Transformer Engine or MSAMP is installed.")
        current_env = setup_fp8_env(args, current_env)
    try: dynamo_backend = DynamoBackend(args.dynamo_backend.upper())
    except ValueError: raise ValueError(f"Unknown dynamo backend: {args.dynamo_backend.upper()}. Choose between {DynamoBackend.list()}.")
    current_env["SAPIENS_ACCELERATOR_DYNAMO_BACKEND"] = dynamo_backend.value
    current_env["SAPIENS_ACCELERATOR_DYNAMO_MODE"] = args.dynamo_mode
    current_env["SAPIENS_ACCELERATOR_DYNAMO_USE_FULLGRAPH"] = str(args.dynamo_use_fullgraph)
    current_env["SAPIENS_ACCELERATOR_DYNAMO_USE_DYNAMIC"] = str(args.dynamo_use_dynamic)
    current_env["OMP_NUM_THREADS"] = str(args.num_cpu_threads_per_process)
    if is_ipex_available():
        current_env["SAPIENS_ACCELERATOR_USE_IPEX"] = str(args.ipex).lower()
        current_env["SAPIENS_ACCELERATOR_USE_XPU"] = str(args.use_xpu).lower()
    if args.enable_cpu_affinity: current_env["SAPIENS_ACCELERATOR_CPU_AFFINITY"] = "1"
    return cmd, current_env
def prepare_multi_gpu_env(args: argparse.Namespace) -> Dict[str, str]:
    num_processes = args.num_processes
    num_machines = args.num_machines
    main_process_ip = args.main_process_ip
    main_process_port = args.main_process_port
    if num_machines > 1:
        args.nproc_per_node = str(num_processes // num_machines)
        args.nnodes = str(num_machines)
        args.node_rank = int(args.machine_rank)
        if getattr(args, "same_network", False):
            args.master_addr = str(main_process_ip)
            args.master_port = str(main_process_port)
        else: args.rdzv_endpoint = f"{main_process_ip}:{main_process_port}"
    else:
        args.nproc_per_node = str(num_processes)
        if main_process_port is not None: args.master_port = str(main_process_port)
    if main_process_port is None: main_process_port = 29500
    need_port_check = num_machines <= 1 or int(args.machine_rank) == 0
    if need_port_check and is_port_in_use(main_process_port): raise ConnectionError(f"Tried to launch distributed communication on port `{main_process_port}`, but another process is utilizing it. Please specify a different port (such as using the `--main_process_port` flag or specifying a different `main_process_port` in your config file) and rerun your script. To automatically use the next open port (on a single node), you can set this to `0`.")
    if args.module and args.no_python: raise ValueError("--module and --no_python cannot be used together")
    elif args.module: args.module = True
    elif args.no_python: args.no_python = True
    current_env = os.environ.copy()
    if args.debug: current_env["SAPIENS_ACCELERATOR_DEBUG_MODE"] = "true"
    gpu_ids = getattr(args, "gpu_ids", "all")
    if gpu_ids != "all" and args.gpu_ids is not None:
        if is_xpu_available(): current_env["ZE_AFFINITY_MASK"] = gpu_ids
        elif is_mlu_available(): current_env["MLU_VISIBLE_DEVICES"] = gpu_ids
        elif is_musa_available(): current_env["MUSA_VISIBLE_DEVICES"] = gpu_ids
        elif is_npu_available(): current_env["ASCEND_RT_VISIBLE_DEVICES"] = gpu_ids
        else: current_env["CUDA_VISIBLE_DEVICES"] = gpu_ids
    mixed_precision = args.mixed_precision.lower()
    try: mixed_precision = PrecisionType(mixed_precision)
    except ValueError: raise ValueError(f"Unknown mixed_precision mode: {mixed_precision}. Choose between {PrecisionType.list()}.")
    current_env["SAPIENS_ACCELERATOR_MIXED_PRECISION"] = str(mixed_precision)
    if args.mixed_precision.lower() == "fp8":
        if not is_fp8_available(): raise RuntimeError("FP8 is not available on this machine. Please ensure that either Transformer Engine or MSAMP is installed.")
        current_env = setup_fp8_env(args, current_env)
    try: dynamo_backend = DynamoBackend(args.dynamo_backend.upper())
    except ValueError: raise ValueError(f"Unknown dynamo backend: {args.dynamo_backend.upper()}. Choose between {DynamoBackend.list()}.")
    current_env["SAPIENS_ACCELERATOR_DYNAMO_BACKEND"] = dynamo_backend.value
    current_env["SAPIENS_ACCELERATOR_DYNAMO_MODE"] = args.dynamo_mode
    current_env["SAPIENS_ACCELERATOR_DYNAMO_USE_FULLGRAPH"] = str(args.dynamo_use_fullgraph)
    current_env["SAPIENS_ACCELERATOR_DYNAMO_USE_DYNAMIC"] = str(args.dynamo_use_dynamic)
    if args.use_fsdp:
        current_env["SAPIENS_ACCELERATOR_USE_FSDP"] = "true"
        if args.fsdp_cpu_ram_efficient_loading and not args.fsdp_sync_module_states: raise ValueError("When using `--fsdp_cpu_ram_efficient_loading` set `--fsdp_sync_module_states` to `True`")
        current_env["FSDP_SHARDING_STRATEGY"] = str(args.fsdp_sharding_strategy)
        current_env["FSDP_OFFLOAD_PARAMS"] = str(args.fsdp_offload_params).lower()
        current_env["FSDP_MIN_NUM_PARAMS"] = str(args.fsdp_min_num_params)
        if args.fsdp_auto_wrap_policy is not None: current_env["FSDP_AUTO_WRAP_POLICY"] = str(args.fsdp_auto_wrap_policy)
        if args.fsdp_transformer_layer_cls_to_wrap is not None: current_env["FSDP_TRANSFORMER_CLS_TO_WRAP"] = str(args.fsdp_transformer_layer_cls_to_wrap)
        if args.fsdp_backward_prefetch is not None: current_env["FSDP_BACKWARD_PREFETCH"] = str(args.fsdp_backward_prefetch)
        if args.fsdp_state_dict_type is not None: current_env["FSDP_STATE_DICT_TYPE"] = str(args.fsdp_state_dict_type)
        current_env["FSDP_FORWARD_PREFETCH"] = str(args.fsdp_forward_prefetch).lower()
        current_env["FSDP_USE_ORIG_PARAMS"] = str(args.fsdp_use_orig_params).lower()
        current_env["FSDP_CPU_RAM_EFFICIENT_LOADING"] = str(args.fsdp_cpu_ram_efficient_loading).lower()
        current_env["FSDP_SYNC_MODULE_STATES"] = str(args.fsdp_sync_module_states).lower()
        current_env["FSDP_ACTIVATION_CHECKPOINTING"] = str(args.fsdp_activation_checkpointing).lower()
    if args.use_megatron_lm:
        prefix = "MEGATRON_LM_"
        current_env["SAPIENS_ACCELERATOR_USE_MEGATRON_LM"] = "true"
        current_env[prefix + "TP_DEGREE"] = str(args.megatron_lm_tp_degree)
        current_env[prefix + "PP_DEGREE"] = str(args.megatron_lm_pp_degree)
        current_env[prefix + "GRADIENT_CLIPPING"] = str(args.megatron_lm_gradient_clipping)
        if args.megatron_lm_num_micro_batches is not None: current_env[prefix + "NUM_MICRO_BATCHES"] = str(args.megatron_lm_num_micro_batches)
        if args.megatron_lm_sequence_parallelism is not None: current_env[prefix + "SEQUENCE_PARALLELISM"] = str(args.megatron_lm_sequence_parallelism)
        if args.megatron_lm_recompute_activations is not None: current_env[prefix + "RECOMPUTE_ACTIVATIONS"] = str(args.megatron_lm_recompute_activations)
        if args.megatron_lm_use_distributed_optimizer is not None: current_env[prefix + "USE_DISTRIBUTED_OPTIMIZER"] = str(args.megatron_lm_use_distributed_optimizer)
    current_env["OMP_NUM_THREADS"] = str(args.num_cpu_threads_per_process)
    if args.enable_cpu_affinity: current_env["SAPIENS_ACCELERATOR_CPU_AFFINITY"] = "1"
    return current_env
def prepare_deepspeed_cmd_env(args: argparse.Namespace) -> Tuple[List[str], Dict[str, str]]:
    num_processes = args.num_processes
    num_machines = args.num_machines
    main_process_ip = args.main_process_ip
    main_process_port = args.main_process_port
    cmd = None
    if args.deepspeed_multinode_launcher is None: args.deepspeed_multinode_launcher = DEEPSPEED_MULTINODE_LAUNCHERS[0]
    if num_machines > 1 and args.deepspeed_multinode_launcher != DEEPSPEED_MULTINODE_LAUNCHERS[1]:
        cmd = ["deepspeed", "--no_local_rank"]
        cmd.extend(["--hostfile", str(args.deepspeed_hostfile), "--launcher", str(args.deepspeed_multinode_launcher)])
        if args.deepspeed_exclusion_filter is not None: cmd.extend(["--exclude", str(args.deepspeed_exclusion_filter)])
        elif args.deepspeed_inclusion_filter is not None: cmd.extend(["--include", str(args.deepspeed_inclusion_filter)])
        else: cmd.extend(["--num_gpus", str(args.num_processes // args.num_machines)])
        if main_process_ip: cmd.extend(["--master_addr", str(main_process_ip)])
        cmd.extend(["--master_port", str(main_process_port)])
        if args.module and args.no_python: raise ValueError("--module and --no_python cannot be used together")
        elif args.module: cmd.append("--module")
        elif args.no_python: cmd.append("--no_python")
        cmd.append(args.training_script)
        cmd.extend(args.training_script_args)
    elif num_machines > 1 and args.deepspeed_multinode_launcher == DEEPSPEED_MULTINODE_LAUNCHERS[1]:
        args.nproc_per_node = str(num_processes // num_machines)
        args.nnodes = str(num_machines)
        args.node_rank = int(args.machine_rank)
        if getattr(args, "same_network", False):
            args.master_addr = str(main_process_ip)
            args.master_port = str(main_process_port)
        else: args.rdzv_endpoint = f"{main_process_ip}:{main_process_port}"
    else:
        args.nproc_per_node = str(num_processes)
        if main_process_port is not None: args.master_port = str(main_process_port)
    if main_process_port is None: main_process_port = 29500
    need_port_check = num_machines <= 1 or int(args.machine_rank) == 0
    if need_port_check and is_port_in_use(main_process_port): raise ConnectionError(f"Tried to launch distributed communication on port `{main_process_port}`, but another process is utilizing it. Please specify a different port (such as using the `--main_process_port` flag or specifying a different `main_process_port` in your config file) and rerun your script. To automatically use the next open port (on a single node), you can set this to `0`.")
    if args.module and args.no_python: raise ValueError("--module and --no_python cannot be used together")
    elif args.module: args.module = True
    elif args.no_python: args.no_python = True
    current_env = os.environ.copy()
    if args.debug: current_env["SAPIENS_ACCELERATOR_DEBUG_MODE"] = "true"
    gpu_ids = getattr(args, "gpu_ids", "all")
    if gpu_ids != "all" and args.gpu_ids is not None:
        if is_xpu_available(): current_env["ZE_AFFINITY_MASK"] = gpu_ids
        elif is_mlu_available(): current_env["MLU_VISIBLE_DEVICES"] = gpu_ids
        elif is_musa_available(): current_env["MUSA_VISIBLE_DEVICES"] = gpu_ids
        elif is_npu_available(): current_env["ASCEND_RT_VISIBLE_DEVICES"] = gpu_ids
        else: current_env["CUDA_VISIBLE_DEVICES"] = gpu_ids
    try: mixed_precision = PrecisionType(args.mixed_precision.lower())
    except ValueError: raise ValueError(f"Unknown mixed_precision mode: {args.mixed_precision.lower()}. Choose between {PrecisionType.list()}.")
    current_env["PYTHONPATH"] = env_var_path_add("PYTHONPATH", os.path.abspath("."))
    current_env["SAPIENS_ACCELERATOR_MIXED_PRECISION"] = str(mixed_precision)
    if args.mixed_precision.lower() == "fp8":
        if not is_fp8_available(): raise RuntimeError("FP8 is not available on this machine. Please ensure that either Transformer Engine or MSAMP is installed.")
        current_env = setup_fp8_env(args, current_env)
    current_env["SAPIENS_ACCELERATOR_CONFIG_DS_FIELDS"] = str(args.deepspeed_fields_from_sapiens_accelerator_config).lower()
    current_env["SAPIENS_ACCELERATOR_USE_DEEPSPEED"] = "true"
    if args.zero_stage is not None: current_env["SAPIENS_ACCELERATOR_DEEPSPEED_ZERO_STAGE"] = str(args.zero_stage)
    if args.gradient_accumulation_steps is not None: current_env["SAPIENS_ACCELERATOR_GRADIENT_ACCUMULATION_STEPS"] = str(args.gradient_accumulation_steps)
    if args.gradient_clipping is not None: current_env["SAPIENS_ACCELERATOR_GRADIENT_CLIPPING"] = str(args.gradient_clipping).lower()
    if args.offload_optimizer_device is not None: current_env["SAPIENS_ACCELERATOR_DEEPSPEED_OFFLOAD_OPTIMIZER_DEVICE"] = str(args.offload_optimizer_device).lower()
    if args.offload_param_device is not None: current_env["SAPIENS_ACCELERATOR_DEEPSPEED_OFFLOAD_PARAM_DEVICE"] = str(args.offload_param_device).lower()
    if args.zero3_init_flag is not None: current_env["SAPIENS_ACCELERATOR_DEEPSPEED_ZERO3_INIT"] = str(args.zero3_init_flag).lower()
    if args.zero3_save_16bit_model is not None: current_env["SAPIENS_ACCELERATOR_DEEPSPEED_ZERO3_SAVE_16BIT_MODEL"] = str(args.zero3_save_16bit_model).lower()
    if args.deepspeed_config_file is not None: current_env["SAPIENS_ACCELERATOR_DEEPSPEED_CONFIG_FILE"] = str(args.deepspeed_config_file)
    if args.enable_cpu_affinity: current_env["SAPIENS_ACCELERATOR_CPU_AFFINITY"] = "1"
    if args.deepspeed_moe_layer_cls_names is not None: current_env["SAPIENS_ACCELERATOR_DEEPSPEED_MOE_LAYER_CLS_NAMES"] = str(args.deepspeed_moe_layer_cls_names)
    return cmd, current_env
def prepare_tpu(args: argparse.Namespace, current_env: Dict[str, str], pod: bool = False) -> Tuple[argparse.Namespace, Dict[str, str]]:
    if args.mixed_precision == "bf16" and is_torch_xla_available(check_is_tpu=True):
        if args.downcast_bf16: current_env["XLA_DOWNCAST_BF16"] = "1"
        else: current_env["XLA_USE_BF16"] = "1"
    if args.debug: current_env["SAPIENS_ACCELERATOR_DEBUG_MODE"] = "true"
    if pod:
        args.vm = args.tpu_vm
        args.tpu = args.tpu_name
    return args, current_env
def _convert_nargs_to_dict(nargs: List[str]) -> Dict[str, str]:
    if len(nargs) < 0: return {}
    def _infer_type(s):
        try:
            s = float(s)
            if s // 1 == s: return int(s)
            return s
        except ValueError: return s
    parser = argparse.ArgumentParser()
    _, unknown = parser.parse_known_args(nargs)
    for index, argument in enumerate(unknown):
        if argument.startswith(("-", "--")):
            action = None
            if index + 1 < len(unknown):
                if unknown[index + 1].startswith(("-", "--")): raise ValueError("SageMaker doesn’t support argparse actions for `store_true` or `store_false`. Please define explicit types")
            else: raise ValueError("SageMaker doesn’t support argparse actions for `store_true` or `store_false`. Please define explicit types")
            if action is None: parser.add_argument(argument, type=_infer_type)
            else: parser.add_argument(argument, action=action)
    return {key: (literal_eval(value) if value in ("True", "False") else value) for key, value in parser.parse_args(nargs).__dict__.items()}
def prepare_sagemager_args_inputs(sagemaker_config: SageMakerConfig, args: argparse.Namespace) -> Tuple[argparse.Namespace, Dict[str, Any]]:
    os.environ["AWS_DEFAULT_REGION"] = sagemaker_config.region
    if sagemaker_config.profile is not None: os.environ["AWS_PROFILE"] = sagemaker_config.profile
    elif args.aws_access_key_id is not None and args.aws_secret_access_key is not None:
        os.environ["AWS_ACCESS_KEY_ID"] = args.aws_access_key_id
        os.environ["AWS_SECRET_ACCESS_KEY"] = args.aws_secret_access_key
    else: raise OSError("You need to provide an aws_access_key_id and aws_secret_access_key when not using aws_profile")
    source_dir = os.path.dirname(args.training_script)
    if not source_dir: source_dir = "."
    entry_point = os.path.basename(args.training_script)
    if not entry_point.endswith(".py"): raise ValueError(f'Your training script should be a python script and not "{entry_point}"')
    hyperparameters = _convert_nargs_to_dict(args.training_script_args)
    try: mixed_precision = PrecisionType(args.mixed_precision.lower())
    except ValueError: raise ValueError(f"Unknown mixed_precision mode: {args.mixed_precision.lower()}. Choose between {PrecisionType.list()}.")
    try: dynamo_backend = DynamoBackend(args.dynamo_backend.upper())
    except ValueError: raise ValueError(f"Unknown dynamo backend: {args.dynamo_backend.upper()}. Choose between {DynamoBackend.list()}.")
    environment = {"SAPIENS_ACCELERATOR_USE_SAGEMAKER": "true", "SAPIENS_ACCELERATOR_MIXED_PRECISION": str(mixed_precision), "SAPIENS_ACCELERATOR_DYNAMO_BACKEND": dynamo_backend.value,
    "SAPIENS_ACCELERATOR_DYNAMO_MODE": args.dynamo_mode, "SAPIENS_ACCELERATOR_DYNAMO_USE_FULLGRAPH": str(args.dynamo_use_fullgraph), "SAPIENS_ACCELERATOR_DYNAMO_USE_DYNAMIC": str(args.dynamo_use_dynamic),
    "SAPIENS_ACCELERATOR_SAGEMAKER_DISTRIBUTED_TYPE": sagemaker_config.distributed_type.value}
    if args.mixed_precision.lower() == "fp8":
        if not is_fp8_available(): raise RuntimeError("FP8 is not available on this machine. Please ensure that either Transformer Engine or MSAMP is installed.")
        environment = setup_fp8_env(args, environment)
    distribution = None
    if sagemaker_config.distributed_type == SageMakerDistributedType.DATA_PARALLEL: distribution = {"smdistributed": {"dataparallel": {"enabled": True}}}
    sagemaker_inputs = None
    if sagemaker_config.sagemaker_inputs_file is not None:
        print(f"Loading SageMaker Inputs from {sagemaker_config.sagemaker_inputs_file} file")
        sagemaker_inputs = {}
        with open(sagemaker_config.sagemaker_inputs_file) as file:
            for i, line in enumerate(file):
                if i == 0: continue
                l = line.split("\t")
                sagemaker_inputs[l[0]] = l[1].strip()
    sagemaker_metrics = None
    if sagemaker_config.sagemaker_metrics_file is not None:
        print(f"Loading SageMaker Metrics from {sagemaker_config.sagemaker_metrics_file} file")
        sagemaker_metrics = []
        with open(sagemaker_config.sagemaker_metrics_file) as file:
            for i, line in enumerate(file):
                if i == 0: continue
                l = line.split("\t")
                metric_dict = {"Name": l[0], "Regex": l[1].strip()}
                sagemaker_metrics.append(metric_dict)
    args = {"image_uri": sagemaker_config.image_uri, "entry_point": entry_point, "source_dir": source_dir, "role": sagemaker_config.iam_role_name,
    "transformers_version": sagemaker_config.transformers_version, "pytorch_version": sagemaker_config.pytorch_version, "py_version": sagemaker_config.py_version,
    "base_job_name": sagemaker_config.base_job_name, "instance_count": sagemaker_config.num_machines, "instance_type": sagemaker_config.ec2_instance_type,
    "debugger_hook_config": False, "distribution": distribution, "hyperparameters": hyperparameters, "environment": environment, "metric_definitions": sagemaker_metrics}
    if sagemaker_config.additional_args is not None: args = merge_dicts(sagemaker_config.additional_args, args)
    return args, sagemaker_inputs
def env_var_path_add(env_var_name, path_to_add):
    paths = [p for p in os.environ.get(env_var_name, "").split(":") if len(p) > 0]
    paths.append(str(path_to_add))
    return ":".join(paths)
class PrepareForLaunch:
    def __init__(self, launcher, distributed_type="NO", debug=False):
        self.launcher = launcher
        self.distributed_type = DistributedType(distributed_type)
        self.debug = debug
    def __call__(self, index, *args):
        if self.debug:
            world_size = int(os.environ.get("WORLD_SIZE"))
            rdv_file = os.environ.get("SAPIENS_ACCELERATOR_DEBUG_RDV_FILE")
            torch.distributed.init_process_group("gloo", rank=index, store=torch.distributed.FileStore(rdv_file, world_size), world_size=world_size)
        elif self.distributed_type in (DistributedType.MULTI_GPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_NPU,
        DistributedType.MULTI_XPU, DistributedType.MULTI_CPU):
            os.environ["LOCAL_RANK"] = str(index)
            nproc = int(os.environ.get("NPROC", 1))
            node_rank = int(os.environ.get("NODE_RANK", 0))
            os.environ["RANK"] = str(nproc * node_rank + index)
        os.environ["FORK_LAUNCHED"] = str(1)
        self.launcher(*args)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from huggingface_hub import model_info
from huggingface_hub.utils import GatedRepoError, RepositoryNotFoundError
from sapiens_accelerator import init_empty_weights
from sapiens_accelerator.commands.utils import CustomArgumentParser
from sapiens_accelerator.utils import (calculate_maximum_sizes, convert_bytes, is_timm_available, is_transformers_available)
if is_transformers_available():
    import transformers
    from transformers import AutoConfig, AutoModel
if is_timm_available(): import timm
def verify_on_hub(repo: str, token: str = None):
    try: return model_info(repo, token=token)
    except (OSError, GatedRepoError): return "gated"
    except RepositoryNotFoundError: return "repo"
def check_has_model(error):
    if is_timm_available() and isinstance(error, RuntimeError) and "Unknown model" in error.args[0]: return "timm"
    elif (is_transformers_available() and isinstance(error, OSError) and "does not appear to have a file named" in error.args[0]): return "transformers"
    else: return "unknown"
def create_empty_model(model_name: str, library_name: str, trust_remote_code: bool = False, access_token: str = None):
    model_info = verify_on_hub(model_name, access_token)
    if model_info == "gated": raise GatedRepoError(f"Repo for model `{model_name}` is gated. You must be authenticated to access it. Please run `huggingface-cli login`.")
    elif model_info == "repo": raise RepositoryNotFoundError(f"Repo for model `{model_name}` does not exist on the Hub. If you are trying to access a private repo, make sure you are authenticated via `huggingface-cli login` and have access.")
    if library_name is None:
        library_name = getattr(model_info, "library_name", False)
        if not library_name: raise ValueError(f"Model `{model_name}` does not have any library metadata on the Hub, please manually pass in a `--library_name` to use (such as `transformers`)")
    if library_name == "transformers":
        if not is_transformers_available(): raise ImportError(f"To check `{model_name}`, `transformers` must be installed. Please install it via `pip install transformers`")
        if model_info.config is None: raise RuntimeError(f"Tried to load `{model_name}` with `transformers` but it does not have any metadata.")
        auto_map = model_info.config.get("auto_map", False)
        config = AutoConfig.from_pretrained(model_name, trust_remote_code=trust_remote_code, token=access_token)
        with init_empty_weights():
            constructor = AutoModel
            if isinstance(auto_map, dict):
                value = None
                for key in auto_map.keys():
                    if key.startswith("AutoModelFor"):
                        value = key
                        break
                if value is not None: constructor = getattr(transformers, value)
            model = constructor.from_config(config, trust_remote_code=trust_remote_code)
    elif library_name == "timm":
        if not is_timm_available(): raise ImportError(f"To check `{model_name}`, `timm` must be installed. Please install it via `pip install timm`")
        with init_empty_weights(): model = timm.create_model(model_name, pretrained=False)
    else: raise ValueError(f"Library `{library_name}` is not supported yet, please open an issue on GitHub for us to add support.")
    return model
def create_ascii_table(headers: list, rows: list, title: str):
    sep_char, in_between = "│", "─"
    column_widths = []
    for i in range(len(headers)):
        column_values = [row[i] for row in rows] + [headers[i]]
        max_column_width = max(len(value) for value in column_values)
        column_widths.append(max_column_width)
    formats = [f"%{column_widths[i]}s" for i in range(len(rows[0]))]
    pattern = f"{sep_char}{sep_char.join(formats)}{sep_char}"
    diff = 0
    def make_row(left_char, middle_char, right_char): return f"{left_char}{middle_char.join([in_between * n for n in column_widths])}{in_between * diff}{right_char}"
    separator = make_row("├", "┼", "┤")
    if len(title) > sum(column_widths):
        diff = abs(len(title) - len(separator))
        column_widths[-1] += diff
    separator = make_row("├", "┼", "┤")
    initial_rows = [make_row("┌", in_between, "┐"), f"{sep_char}{title.center(len(separator) - 2)}{sep_char}", make_row("├", "┬", "┤")]
    table = "\n".join(initial_rows) + "\n"
    column_widths[-1] += diff
    centered_line = [text.center(column_widths[i]) for i, text in enumerate(headers)]
    table += f"{pattern % tuple(centered_line)}\n{separator}\n"
    for i, line in enumerate(rows):
        centered_line = [t.center(column_widths[i]) for i, t in enumerate(line)]
        table += f"{pattern % tuple(centered_line)}\n"
    table += f'└{"┴".join([in_between * n for n in column_widths])}┘'
    return table
def estimate_command_parser(subparsers=None):
    if subparsers is not None: parser = subparsers.add_parser("estimate-memory")
    else: parser = CustomArgumentParser(description="Model size estimator for fitting a model onto CUDA memory.")
    parser.add_argument("model_name", type=str, help="")
    parser.add_argument("--library_name", type=str, help="The library the model has an integration with, such as `transformers`, needed only if this information is not stored on the Hub.", choices=["timm", "transformers"])
    parser.add_argument("--dtypes", type=str, nargs="+", default=["float32", "float16", "int8", "int4"], help="The dtypes to use for the model, must be one (or many) of `float32`, `float16`, `int8`, and `int4`", choices=["float32", "float16", "int8", "int4"])
    parser.add_argument("--trust_remote_code", action="store_true", help="Whether or not to allow for custom models defined on the Hub in their own modeling files. This flag should only be used for repositories you trust and in which you have read the code, as it will execute code present on the Hub on your local machine.", default=False)
    if subparsers is not None: parser.set_defaults(func=estimate_command)
    return parser
def estimate_training_usage(bytes: int, mixed_precision: str, msamp_config: str = None) -> dict:
    memory_sizes = {"model": -1, "optimizer": -1, "gradients": -1, "step": -1}
    fp32_size = bytes
    fp16_size = bytes // 2
    if mixed_precision == "float32":
        memory_sizes["model"] = fp32_size
        memory_sizes["gradients"] = fp32_size
        memory_sizes["optimizer"] = fp32_size * 2
        memory_sizes["step"] = fp32_size * 4
    elif mixed_precision in ("float16", "bfloat16") or (mixed_precision == "fp8" and msamp_config is None):
        memory_sizes["model"] = fp32_size
        memory_sizes["gradients"] = fp32_size + fp16_size
        memory_sizes["optimizer"] = fp32_size * 2
        memory_sizes["step"] = memory_sizes["optimizer"]
    return memory_sizes
def gather_data(args):
    try: model = create_empty_model(args.model_name, library_name=args.library_name, trust_remote_code=args.trust_remote_code)
    except (RuntimeError, OSError) as e:
        library = check_has_model(e)
        if library != "unknown": raise RuntimeError(f"Tried to load `{args.model_name}` with `{library}` but a possible model to load was not found inside the repo.")
        raise e
    total_size, largest_layer = calculate_maximum_sizes(model)
    data = []
    for dtype in args.dtypes:
        dtype_total_size = total_size
        dtype_largest_layer = largest_layer[0]
        dtype_training_size = estimate_training_usage(dtype_total_size, dtype)
        if dtype == "float16":
            dtype_total_size /= 2
            dtype_largest_layer /= 2
        elif dtype == "int8":
            dtype_total_size /= 4
            dtype_largest_layer /= 4
        elif dtype == "int4":
            dtype_total_size /= 8
            dtype_largest_layer /= 8
        data.append([dtype, dtype_largest_layer, dtype_total_size, dtype_training_size])
    return data
def estimate_command(args): pass
def main():
    parser = estimate_command_parser()
    args = parser.parse_args()
    estimate_command(args)
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import json
import os
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Union
import yaml
from ...utils import ComputeEnvironment, DistributedType, SageMakerDistributedType
from ...utils.constants import SAGEMAKER_PYTHON_VERSION, SAGEMAKER_PYTORCH_VERSION, SAGEMAKER_TRANSFORMERS_VERSION
hf_cache_home = os.path.expanduser(os.environ.get("HF_HOME", os.path.join(os.environ.get("XDG_CACHE_HOME", "~/.cache"), "huggingface")))
cache_dir = os.path.join(hf_cache_home, "sapiens_accelerator")
default_json_config_file = os.path.join(cache_dir, "default_config.yaml")
default_yaml_config_file = os.path.join(cache_dir, "default_config.yaml")
if os.path.isfile(default_yaml_config_file) or not os.path.isfile(default_json_config_file): default_config_file = default_yaml_config_file
else: default_config_file = default_json_config_file
def load_config_from_file(config_file):
    if config_file is not None:
        if not os.path.isfile(config_file): raise FileNotFoundError(f"The passed configuration file `{config_file}` does not exist. Please pass an existing file to `sapiens_accelerator launch`, or use the default one created through `sapiens_accelerator config` and run `sapiens_accelerator launch` without the `--config_file` argument.")
    else: config_file = default_config_file
    with open(config_file, encoding="utf-8") as f:
        if config_file.endswith(".json"):
            if (json.load(f).get("compute_environment", ComputeEnvironment.LOCAL_MACHINE) == ComputeEnvironment.LOCAL_MACHINE): config_class = ClusterConfig
            else: config_class = SageMakerConfig
            return config_class.from_json_file(json_file=config_file)
        else:
            if (yaml.safe_load(f).get("compute_environment", ComputeEnvironment.LOCAL_MACHINE) == ComputeEnvironment.LOCAL_MACHINE): config_class = ClusterConfig
            else: config_class = SageMakerConfig
            return config_class.from_yaml_file(yaml_file=config_file)
@dataclass
class BaseConfig:
    compute_environment: ComputeEnvironment
    distributed_type: Union[DistributedType, SageMakerDistributedType]
    mixed_precision: str
    use_cpu: bool
    debug: bool
    def to_dict(self):
        result = self.__dict__
        def _convert_enums(value):
            if isinstance(value, Enum): return value.value
            if isinstance(value, dict):
                if not bool(value): return None
                for key1, value1 in value.items(): value[key1] = _convert_enums(value1)
            return value
        for key, value in result.items(): result[key] = _convert_enums(value)
        result = {k: v for k, v in result.items() if v is not None}
        return result
    @staticmethod
    def process_config(config_dict):
        if "compute_environment" not in config_dict: config_dict["compute_environment"] = ComputeEnvironment.LOCAL_MACHINE
        if "distributed_type" not in config_dict: raise ValueError("A `distributed_type` must be specified in the config file.")
        if "num_processes" not in config_dict and config_dict["distributed_type"] == DistributedType.NO: config_dict["num_processes"] = 1
        if "mixed_precision" not in config_dict: config_dict["mixed_precision"] = "fp16" if ("fp16" in config_dict and config_dict["fp16"]) else None
        if "fp16" in config_dict: del config_dict["fp16"]
        if "dynamo_backend" in config_dict:
            dynamo_backend = config_dict.pop("dynamo_backend")
            config_dict["dynamo_config"] = {} if dynamo_backend == "NO" else {"dynamo_backend": dynamo_backend}
        if "use_cpu" not in config_dict: config_dict["use_cpu"] = False
        if "debug" not in config_dict: config_dict["debug"] = False
        if "enable_cpu_affinity" not in config_dict: config_dict["enable_cpu_affinity"] = False
        return config_dict
    @classmethod
    def from_json_file(cls, json_file=None):
        json_file = default_json_config_file if json_file is None else json_file
        with open(json_file, encoding="utf-8") as f: config_dict = json.load(f)
        config_dict = cls.process_config(config_dict)
        extra_keys = sorted(set(config_dict.keys()) - set(cls.__dataclass_fields__.keys()))
        if len(extra_keys) > 0: raise ValueError(f"The config file at {json_file} had unknown keys ({extra_keys}), please try upgrading your `sapiens_accelerator` version or fix (and potentially remove) these keys from your config file.")
        return cls(**config_dict)
    def to_json_file(self, json_file):
        with open(json_file, "w", encoding="utf-8") as f:
            content = json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"
            f.write(content)
    @classmethod
    def from_yaml_file(cls, yaml_file=None):
        yaml_file = default_yaml_config_file if yaml_file is None else yaml_file
        with open(yaml_file, encoding="utf-8") as f: config_dict = yaml.safe_load(f)
        config_dict = cls.process_config(config_dict)
        extra_keys = sorted(set(config_dict.keys()) - set(cls.__dataclass_fields__.keys()))
        if len(extra_keys) > 0: raise ValueError(f"The config file at {yaml_file} had unknown keys ({extra_keys}), please try upgrading your `sapiens_accelerator` version or fix (and potentially remove) these keys from your config file.")
        return cls(**config_dict)
    def to_yaml_file(self, yaml_file):
        with open(yaml_file, "w", encoding="utf-8") as f: yaml.safe_dump(self.to_dict(), f)
    def __post_init__(self):
        if isinstance(self.compute_environment, str): self.compute_environment = ComputeEnvironment(self.compute_environment)
        if isinstance(self.distributed_type, str):
            if self.compute_environment == ComputeEnvironment.AMAZON_SAGEMAKER: self.distributed_type = SageMakerDistributedType(self.distributed_type)
            else: self.distributed_type = DistributedType(self.distributed_type)
        if getattr(self, "dynamo_config", None) is None: self.dynamo_config = {}
@dataclass
class ClusterConfig(BaseConfig):
    num_processes: int = -1
    machine_rank: int = 0
    num_machines: int = 1
    gpu_ids: Optional[str] = None
    main_process_ip: Optional[str] = None
    main_process_port: Optional[int] = None
    rdzv_backend: Optional[str] = "static"
    same_network: Optional[bool] = False
    main_training_function: str = "main"
    enable_cpu_affinity: bool = False
    fp8_config: dict = None
    deepspeed_config: dict = None
    fsdp_config: dict = None
    megatron_lm_config: dict = None
    ipex_config: dict = None
    mpirun_config: dict = None
    downcast_bf16: bool = False
    tpu_name: str = None
    tpu_zone: str = None
    tpu_use_cluster: bool = False
    tpu_use_sudo: bool = False
    command_file: str = None
    commands: List[str] = None
    tpu_vm: List[str] = None
    tpu_env: List[str] = None
    dynamo_config: dict = None
    def __post_init__(self):
        if self.deepspeed_config is None: self.deepspeed_config = {}
        if self.fsdp_config is None: self.fsdp_config = {}
        if self.megatron_lm_config is None: self.megatron_lm_config = {}
        if self.ipex_config is None: self.ipex_config = {}
        if self.mpirun_config is None: self.mpirun_config = {}
        if self.fp8_config is None: self.fp8_config = {}
        return super().__post_init__()
@dataclass
class SageMakerConfig(BaseConfig):
    ec2_instance_type: str
    iam_role_name: str
    image_uri: Optional[str] = None
    profile: Optional[str] = None
    region: str = "us-east-1"
    num_machines: int = 1
    gpu_ids: str = "all"
    base_job_name: str = f"sapiens_accelerator-sagemaker-{num_machines}"
    pytorch_version: str = SAGEMAKER_PYTORCH_VERSION
    transformers_version: str = SAGEMAKER_TRANSFORMERS_VERSION
    py_version: str = SAGEMAKER_PYTHON_VERSION
    sagemaker_inputs_file: str = None
    sagemaker_metrics_file: str = None
    additional_args: dict = None
    dynamo_config: dict = None
    enable_cpu_affinity: bool = False
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import contextlib
import io
import math
import time
from copy import deepcopy
from pathlib import Path
import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset
from sapiens_accelerator import Accelerator
from sapiens_accelerator.data_loader import SeedableRandomSampler, prepare_data_loader
from sapiens_accelerator.state import AcceleratorState
from sapiens_accelerator.test_utils import RegressionDataset, are_the_same_tensors
from sapiens_accelerator.utils import (DataLoaderConfiguration, DistributedType, gather, is_bf16_available, is_datasets_available, is_ipex_available,
is_mlu_available, is_musa_available, is_npu_available, is_pytest_available, is_xpu_available, set_seed, synchronize_rng_states)
if is_xpu_available(): from sapiens_accelerator.test_utils import RegressionModel4XPU as RegressionModel
else: from sapiens_accelerator.test_utils import RegressionModel
def generate_baseline_dataloader(train_set, generator, batch_size, use_seedable_sampler=False):
    if use_seedable_sampler:
        sampler = SeedableRandomSampler(generator=generator, data_source=train_set, num_samples=len(train_set))
        return DataLoader(train_set, batch_size=batch_size, sampler=sampler)
    else: return DataLoader(train_set, batch_size=batch_size, shuffle=True, generator=generator)
def print_main(state): pass
def print_local_main(state): pass
def print_last(state): pass
def print_on(state, process_idx): pass
def process_execution_check():
    accelerator = Accelerator()
    num_processes = accelerator.num_processes
    path = Path("check_main_process_first.txt")
    with accelerator.main_process_first():
        if accelerator.is_main_process:
            time.sleep(0.1)
            with open(path, "a+") as f: f.write("Currently in the main process\n")
        else:
            with open(path, "a+") as f: f.write("Now on another process\n")
    accelerator.wait_for_everyone()
    if accelerator.is_main_process:
        with open(path) as f: text = "".join(f.readlines())
        try:
            assert text.startswith("Currently in the main process\n"), "Main process was not first"
            if num_processes > 1: assert text.endswith("Now on another process\n"), "Main process was not first"
            assert (text.count("Now on another process\n") == accelerator.num_processes - 1), f"Only wrote to file {text.count('Now on another process') + 1} times, not {accelerator.num_processes}"
        except AssertionError:
            path.unlink()
            raise
    if accelerator.is_main_process and path.exists(): path.unlink()
    accelerator.wait_for_everyone()
    f = io.StringIO()
    with contextlib.redirect_stdout(f): accelerator.on_main_process(print_main)(accelerator.state)
    result = f.getvalue().rstrip()
    if accelerator.is_main_process: assert result == "Printing from the main process 0", f"{result} != Printing from the main process 0"
    else: assert f.getvalue().rstrip() == "", f'{result} != ""'
    f.truncate(0)
    f.seek(0)
    with contextlib.redirect_stdout(f): accelerator.on_local_main_process(print_local_main)(accelerator.state)
    if accelerator.is_local_main_process: assert f.getvalue().rstrip() == "Printing from the local main process 0"
    else: assert f.getvalue().rstrip() == ""
    f.truncate(0)
    f.seek(0)
    with contextlib.redirect_stdout(f): accelerator.on_last_process(print_last)(accelerator.state)
    if accelerator.is_last_process: assert f.getvalue().rstrip() == f"Printing from the last process {accelerator.state.num_processes - 1}"
    else: assert f.getvalue().rstrip() == ""
    f.truncate(0)
    f.seek(0)
    for process_idx in range(num_processes):
        with contextlib.redirect_stdout(f): accelerator.on_process(print_on, process_index=process_idx)(accelerator.state, process_idx)
        if accelerator.process_index == process_idx: assert f.getvalue().rstrip() == f"Printing from process {process_idx}: {accelerator.process_index}"
        else: assert f.getvalue().rstrip() == ""
        f.truncate(0)
        f.seek(0)
def init_state_check(): state = AcceleratorState()
def rng_sync_check():
    state = AcceleratorState()
    synchronize_rng_states(["torch"])
    assert are_the_same_tensors(torch.get_rng_state()), "RNG states improperly synchronized on CPU."
    if state.distributed_type == DistributedType.MULTI_GPU:
        synchronize_rng_states(["cuda"])
        assert are_the_same_tensors(torch.cuda.get_rng_state()), "RNG states improperly synchronized on GPU."
    elif state.distributed_type == DistributedType.MULTI_XPU:
        synchronize_rng_states(["xpu"])
        assert are_the_same_tensors(torch.xpu.get_rng_state()), "RNG states improperly synchronized on XPU."
    generator = torch.Generator()
    synchronize_rng_states(["generator"], generator=generator)
    assert are_the_same_tensors(generator.get_state()), "RNG states improperly synchronized in generator."
def dl_preparation_check():
    state = AcceleratorState()
    length = 32 * state.num_processes
    dl = DataLoader(range(length), batch_size=8)
    dl = prepare_data_loader(dl, state.device, state.num_processes, state.process_index, put_on_device=True)
    result = []
    for batch in dl: result.append(gather(batch))
    result = torch.cat(result)
    assert torch.equal(result.cpu(), torch.arange(0, length).long()), "Wrong non-shuffled dataloader result."
    dl = DataLoader(range(length), batch_size=8)
    dl = prepare_data_loader(dl, state.device, state.num_processes, state.process_index, put_on_device=True, split_batches=True)
    result = []
    for batch in dl: result.append(gather(batch))
    result = torch.cat(result)
    assert torch.equal(result.cpu(), torch.arange(0, length).long()), "Wrong non-shuffled dataloader result."
    dl = DataLoader(range(length), batch_size=8, shuffle=True)
    dl = prepare_data_loader(dl, state.device, state.num_processes, state.process_index, put_on_device=True)
    result = []
    for batch in dl: result.append(gather(batch))
    result = torch.cat(result).tolist()
    result.sort()
    assert result == list(range(length)), "Wrong shuffled dataloader result."
    dl = DataLoader(range(length), batch_size=8, shuffle=True)
    dl = prepare_data_loader(dl, state.device, state.num_processes, state.process_index, put_on_device=True, split_batches=True)
    result = []
    for batch in dl: result.append(gather(batch))
    result = torch.cat(result).tolist()
    result.sort()
    assert result == list(range(length)), "Wrong shuffled dataloader result."
def central_dl_preparation_check():
    state = AcceleratorState()
    length = 32 * state.num_processes
    dl = DataLoader(range(length), batch_size=8)
    dl = prepare_data_loader(dl, state.device, state.num_processes, state.process_index, put_on_device=True, dispatch_batches=True)
    result = []
    for batch in dl: result.append(gather(batch))
    result = torch.cat(result)
    assert torch.equal(result.cpu(), torch.arange(0, length).long()), "Wrong non-shuffled dataloader result."
    dl = DataLoader(range(length), batch_size=8)
    dl = prepare_data_loader(dl, state.device, state.num_processes, state.process_index, put_on_device=True, split_batches=True, dispatch_batches=True)
    result = []
    for batch in dl: result.append(gather(batch))
    result = torch.cat(result)
    assert torch.equal(result.cpu(), torch.arange(0, length).long()), "Wrong non-shuffled dataloader result."
    dl = DataLoader(range(length), batch_size=8, shuffle=True)
    dl = prepare_data_loader(dl, state.device, state.num_processes, state.process_index, put_on_device=True, dispatch_batches=True)
    result = []
    for batch in dl: result.append(gather(batch))
    result = torch.cat(result).tolist()
    result.sort()
    assert result == list(range(length)), "Wrong shuffled dataloader result."
    dl = DataLoader(range(length), batch_size=8, shuffle=True)
    dl = prepare_data_loader(dl, state.device, state.num_processes, state.process_index, put_on_device=True, split_batches=True, dispatch_batches=True)
    result = []
    for batch in dl: result.append(gather(batch))
    result = torch.cat(result).tolist()
    result.sort()
    assert result == list(range(length)), "Wrong shuffled dataloader result."
def custom_sampler_check():
    state = AcceleratorState()
    class CustomDataset(Dataset):
        def __init__(self, data): self.data = data
        def __len__(self): return len(self.data)
        def __getitem__(self, index): return self.data[index]
    class CustomBatchSampler:
        def __init__(self, dataset_length: int, batch_size: int, shuffle: bool = True):
            self.batch_size = batch_size
            self.data_index = np.arange(dataset_length)
            self.shuffle = shuffle
        def __iter__(self):
            num_batches = len(self)
            if self.shuffle: index = np.random.permutation(self.data_index)
            else: index = self.data_index
            output = np.array_split(index, num_batches)
            yield from output
        def __len__(self): return math.ceil(len(self.data_index) / self.batch_size)
    dataset = CustomDataset(range(32 * state.num_processes))
    sampler = CustomBatchSampler(len(dataset), batch_size=8)
    dl = DataLoader(dataset, batch_sampler=sampler)
    dl = prepare_data_loader(dl, state.device, state.num_processes, state.process_index)
    if hasattr(dl.batch_sampler, "batch_sampler"): assert isinstance(dl.batch_sampler.batch_sampler, CustomBatchSampler), "Custom sampler was changed after calling `prepare_data_loader`"
    else: assert isinstance(dl.batch_sampler, CustomBatchSampler), "Custom sampler was changed after calling `prepare_data_loader`"
def check_seedable_sampler():
    set_seed(42)
    train_set = RegressionDataset(length=10, seed=42)
    train_dl = DataLoader(train_set, batch_size=2, shuffle=True)
    config = DataLoaderConfiguration(use_seedable_sampler=True)
    accelerator = Accelerator(dataloader_config=config)
    train_dl = accelerator.prepare(train_dl)
    original_items = []
    for _ in range(3):
        for batch in train_dl: original_items.append(batch["x"])
    original_items = torch.cat(original_items)
    set_seed(42)
    train_dl.set_epoch(0)
    new_items = []
    for _ in range(3):
        for batch in train_dl: new_items.append(batch["x"])
    new_items = torch.cat(new_items)
    assert torch.allclose(original_items, new_items), "Did not obtain the same items with the same seed and epoch."
def check_seedable_sampler_in_batch_sampler_shard():
    set_seed(42)
    config = DataLoaderConfiguration(use_seedable_sampler=True)
    accelerator = Accelerator(dataloader_config=config)
    assert accelerator.num_processes > 1, "This test requires more than one process."
    dataloader = DataLoader(list(range(10)), batch_size=1, shuffle=True)
    prepared_data_loader = prepare_data_loader(dataloader=dataloader, use_seedable_sampler=True)
    target_sampler = prepared_data_loader.batch_sampler.batch_sampler.sampler
    assert isinstance(target_sampler, SeedableRandomSampler), "Sampler in BatchSamplerShard is not SeedableRandomSampler."
def mock_training(length, batch_size, generator, use_seedable_sampler=False):
    set_seed(42)
    generator.manual_seed(42)
    train_set = RegressionDataset(length=length, seed=42)
    train_dl = generate_baseline_dataloader(train_set, generator, batch_size, use_seedable_sampler)
    model = RegressionModel()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.1)
    for epoch in range(3):
        for batch in train_dl:
            model.zero_grad()
            output = model(batch["x"])
            loss = torch.nn.functional.mse_loss(output, batch["y"])
            loss.backward()
            optimizer.step()
    return train_set, model
def training_check(use_seedable_sampler=False):
    state = AcceleratorState()
    generator = torch.Generator()
    batch_size = 8
    length = batch_size * 4 * state.num_processes
    train_set, old_model = mock_training(length, batch_size * state.num_processes, generator, use_seedable_sampler)
    assert are_the_same_tensors(old_model.a), "Did not obtain the same model on both processes."
    assert are_the_same_tensors(old_model.b), "Did not obtain the same model on both processes."
    accelerator = Accelerator()
    train_dl = generate_baseline_dataloader(train_set, generator, batch_size, use_seedable_sampler)
    model = RegressionModel()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.1)
    train_dl, model, optimizer = accelerator.prepare(train_dl, model, optimizer)
    set_seed(42)
    generator.manual_seed(42)
    for _ in range(3):
        for batch in train_dl:
            model.zero_grad()
            output = model(batch["x"])
            loss = torch.nn.functional.mse_loss(output, batch["y"])
            accelerator.backward(loss)
            optimizer.step()
    model = accelerator.unwrap_model(model).cpu()
    assert torch.allclose(old_model.a, model.a), "Did not obtain the same model on CPU or distributed training."
    assert torch.allclose(old_model.b, model.b), "Did not obtain the same model on CPU or distributed training."
    dataloader_config = DataLoaderConfiguration(split_batches=True, use_seedable_sampler=use_seedable_sampler)
    accelerator = Accelerator(dataloader_config=dataloader_config)
    train_dl = generate_baseline_dataloader(train_set, generator, batch_size * state.num_processes, use_seedable_sampler)
    model = RegressionModel()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.1)
    train_dl, model, optimizer = accelerator.prepare(train_dl, model, optimizer)
    set_seed(42)
    generator.manual_seed(42)
    for _ in range(3):
        for batch in train_dl:
            model.zero_grad()
            output = model(batch["x"])
            loss = torch.nn.functional.mse_loss(output, batch["y"])
            accelerator.backward(loss)
            optimizer.step()
    model = accelerator.unwrap_model(model).cpu()
    assert torch.allclose(old_model.a, model.a), "Did not obtain the same model on CPU or distributed training."
    assert torch.allclose(old_model.b, model.b), "Did not obtain the same model on CPU or distributed training."
    if torch.cuda.is_available() or is_npu_available() or is_mlu_available() or is_musa_available():
        AcceleratorState._reset_state()
        dataloader_config = DataLoaderConfiguration(use_seedable_sampler=use_seedable_sampler)
        accelerator = Accelerator(mixed_precision="fp16", dataloader_config=dataloader_config)
        train_dl = generate_baseline_dataloader(train_set, generator, batch_size, use_seedable_sampler)
        model = RegressionModel()
        optimizer = torch.optim.SGD(model.parameters(), lr=0.1)
        train_dl, model, optimizer = accelerator.prepare(train_dl, model, optimizer)
        set_seed(42)
        generator.manual_seed(42)
        for _ in range(3):
            for batch in train_dl:
                model.zero_grad()
                output = model(batch["x"])
                loss = torch.nn.functional.mse_loss(output, batch["y"])
                accelerator.backward(loss)
                optimizer.step()
        model = accelerator.unwrap_model(model).cpu()
        assert torch.allclose(old_model.a, model.a), "Did not obtain the same model on CPU or distributed training."
        assert torch.allclose(old_model.b, model.b), "Did not obtain the same model on CPU or distributed training."
    if torch.cuda.is_available():
        AcceleratorState._reset_state()
        accelerator = Accelerator(mixed_precision="fp16")
        model = torch.nn.Linear(2, 4)
        model = accelerator.prepare(model)
        model_with_fp32_wrapper = accelerator.unwrap_model(model, keep_fp32_wrapper=True)
        input_tensor = torch.Tensor([1, 2]).to(dtype=torch.float16, device=accelerator.device)
        output = model_with_fp32_wrapper(input_tensor)
    if is_bf16_available():
        AcceleratorState._reset_state()
        dataloader_config = DataLoaderConfiguration(use_seedable_sampler=use_seedable_sampler)
        accelerator = Accelerator(mixed_precision="bf16", dataloader_config=dataloader_config)
        train_dl = generate_baseline_dataloader(train_set, generator, batch_size, use_seedable_sampler)
        model = RegressionModel()
        optimizer = torch.optim.SGD(model.parameters(), lr=0.1)
        train_dl, model, optimizer = accelerator.prepare(train_dl, model, optimizer)
        set_seed(42)
        generator.manual_seed(42)
        for _ in range(3):
            for batch in train_dl:
                model.zero_grad()
                output = model(batch["x"])
                loss = torch.nn.functional.mse_loss(output, batch["y"])
                accelerator.backward(loss)
                optimizer.step()
        model = accelerator.unwrap_model(model).cpu()
        assert torch.allclose(old_model.a, model.a), "Did not obtain the same model on CPU or distributed training."
        assert torch.allclose(old_model.b, model.b), "Did not obtain the same model on CPU or distributed training."
    if is_ipex_available():
        AcceleratorState._reset_state()
        dataloader_config = DataLoaderConfiguration(use_seedable_sampler=use_seedable_sampler)
        accelerator = Accelerator(mixed_precision="bf16", cpu=True, dataloader_config=dataloader_config)
        train_dl = generate_baseline_dataloader(train_set, generator, batch_size, use_seedable_sampler)
        model = RegressionModel()
        optimizer = torch.optim.SGD(model.parameters(), lr=0.1)
        train_dl, model, optimizer = accelerator.prepare(train_dl, model, optimizer)
        set_seed(42)
        generator.manual_seed(42)
        for _ in range(3):
            for batch in train_dl:
                model.zero_grad()
                output = model(batch["x"])
                loss = torch.nn.functional.mse_loss(output, batch["y"])
                accelerator.backward(loss)
                optimizer.step()
        model = accelerator.unwrap_model(model).cpu()
        assert torch.allclose(old_model.a, model.a), "Did not obtain the same model on CPU or distributed training."
        assert torch.allclose(old_model.b, model.b), "Did not obtain the same model on CPU or distributed training."
    if is_xpu_available():
        AcceleratorState._reset_state()
        dataloader_config = DataLoaderConfiguration(use_seedable_sampler=use_seedable_sampler)
        accelerator = Accelerator(mixed_precision="bf16", cpu=False, dataloader_config=dataloader_config)
        train_dl = generate_baseline_dataloader(train_set, generator, batch_size, use_seedable_sampler)
        model = RegressionModel()
        optimizer = torch.optim.SGD(model.parameters(), lr=0.1)
        train_dl, model, optimizer = accelerator.prepare(train_dl, model, optimizer)
        set_seed(42)
        generator.manual_seed(42)
        for _ in range(3):
            for batch in train_dl:
                model.zero_grad()
                output = model(batch["x"])
                loss = torch.nn.functional.mse_loss(output, batch["y"])
                accelerator.backward(loss)
                optimizer.step()
        model = accelerator.unwrap_model(model).cpu()
        assert torch.allclose(old_model.a, model.a), "Did not obtain the same model on XPU or distributed training."
        assert torch.allclose(old_model.b, model.b), "Did not obtain the same model on XPU or distributed training."
def test_split_between_processes_dataset(datasets_Dataset):
    state = AcceleratorState()
    data = datasets_Dataset.from_list([dict(k=v) for v in range(2 * state.num_processes)])
    with state.split_between_processes(data, apply_padding=False) as results: assert (len(results) == 2), f"Each process did not have two items. Process index: {state.process_index}; Length: {len(results)}"
    data = datasets_Dataset.from_list([dict(k=v) for v in range(2 * state.num_processes - 1)])
    with state.split_between_processes(data, apply_padding=False) as results:
        if state.is_last_process: assert (len(results) == 1), f"Last process did not receive a single item. Process index: {state.process_index}; Length: {len(results)}"
        else: assert (len(results) == 2), f"One of the intermediate processes did not receive two items. Process index: {state.process_index}; Length: {len(results)}"
    data = datasets_Dataset.from_list([dict(k=v) for v in range(2 * state.num_processes - 1)])
    with state.split_between_processes(data, apply_padding=True) as results:
        if state.num_processes == 1: assert (len(results) == 1), f"Single process did not receive a single item. Process index: {state.process_index}; Length: {len(results)}"
        else: assert (len(results) == 2), f"Each process did not have two items. Process index: {state.process_index}; Length: {len(results)}"
    state.wait_for_everyone()
def test_split_between_processes_list():
    state = AcceleratorState()
    data = list(range(0, 2 * state.num_processes))
    with state.split_between_processes(data) as results: assert (len(results) == 2), f"Each process did not have two items. Process index: {state.process_index}; Length: {len(results)}"
    data = list(range(0, (3 * state.num_processes) - 1))
    with state.split_between_processes(data, apply_padding=True) as results:
        if state.is_last_process:
            num_samples_per_device = math.ceil(len(data) / state.num_processes)
            assert (len(results) == num_samples_per_device), f"Last process did not get the extra item(s). Process index: {state.process_index}; Length: {len(results)}"
    state.wait_for_everyone()
def test_split_between_processes_nested_dict():
    state = AcceleratorState()
    a = [1, 2, 3, 4, 5, 6, 7, 8]
    b = ["a", "b", "c", "d", "e", "f", "g", "h"]
    c = torch.tensor([1, 2, 3, 4, 5, 6, 7, 8])
    if state.num_processes in (1, 2, 4):
        data = {"a": a, "b": b, "c": c}
        data_copy = deepcopy(data)
        with state.split_between_processes(data) as results:
            if state.process_index == 0: assert results["a"] == data_copy["a"][: 8 // state.num_processes]
            elif state.num_processes == 2: assert results["a"] == data_copy["a"][4:]
            elif state.process_index == 3: assert results["a"] == data_copy["a"][-2:], f'Expected: {data_copy["a"][-2]}, Actual: {results["a"]}'
            if state.process_index == 0: assert results["b"] == data_copy["b"][: 8 // state.num_processes]
            elif state.num_processes == 2: assert results["b"] == data_copy["b"][4:]
            elif state.process_index == 3: assert results["b"] == data_copy["b"][-2:]
            if state.process_index == 0: assert torch.allclose(results["c"], data_copy["c"][: 8 // state.num_processes]), f"Did not obtain expected values on process 0, expected `{data['c'][:8 // state.num_processes]}`, received: {results['c']}"
            elif state.num_processes == 2: assert torch.allclose(results["c"], data_copy["c"][4:]), f"Did not obtain expected values on process 2, expected `{data['c'][4:]}`, received: {results['c']}"
            elif state.process_index == 3: assert torch.allclose(results["c"], data_copy["c"][-2:]), f"Did not obtain expected values on process 4, expected `{data['c'][-2:]}`, received: {results['c']}"
    state.wait_for_everyone()
def test_split_between_processes_tensor():
    state = AcceleratorState()
    if state.num_processes > 1:
        data = torch.tensor([[0, 1, 2, 3], [4, 5, 6, 7]]).to(state.device)
        with state.split_between_processes(data) as results:
            if state.process_index == 0: assert torch.allclose(results, torch.tensor([0, 1, 2, 3]).to(state.device))
            else: assert torch.allclose(results, torch.tensor([4, 5, 6, 7]).to(state.device))
    state.wait_for_everyone()
def test_split_between_processes_evenly():
    state = AcceleratorState()
    if state.num_processes in (1, 2, 4, 8):
        data = list(range(17))
        num_samples_per_process = len(data) // state.num_processes
        num_extras = len(data) % state.num_processes
        with state.split_between_processes(data) as results:
            if state.process_index < num_extras: assert (len(results) == num_samples_per_process + 1), f"Each Process should have even elements. Expected: {num_samples_per_process + 1}, Actual: {len(results)}"
            else: assert (len(results) == num_samples_per_process), f"Each Process should have even elements. Expected: {num_samples_per_process}, Actual: {len(results)}"
    state.wait_for_everyone()
def test_trigger():
    accelerator = Accelerator()
    assert accelerator.check_trigger() is False
    if accelerator.is_main_process: accelerator.set_trigger()
    assert accelerator.check_trigger() is True
    assert accelerator.check_trigger() is False
def test_reinstantiated_state():
    import pytest
    AcceleratorState._reset_state()
    simple_model = torch.nn.Linear(1, 1)
    accelerator = Accelerator()
    AcceleratorState._reset_state()
    with pytest.raises(AttributeError) as cm: accelerator.prepare(simple_model)
    assert "`AcceleratorState` object has no attribute" in str(cm.value.args[0])
    assert "This happens if `AcceleratorState._reset_state()`" in str(cm.value.args[0])
def main():
    accelerator = Accelerator()
    state = accelerator.state
    init_state_check()
    state.wait_for_everyone()
    if state.distributed_type == DistributedType.MULTI_GPU: num_processes_per_node = torch.cuda.device_count()
    else: num_processes_per_node = state.num_processes
    if num_processes_per_node == state.num_processes:
        process_execution_check()
        test_split_between_processes_list()
        test_split_between_processes_nested_dict()
        test_split_between_processes_tensor()
        test_split_between_processes_evenly()
        if is_datasets_available():
            from datasets import Dataset as datasets_Dataset
            test_split_between_processes_dataset(datasets_Dataset)
    rng_sync_check()
    dl_preparation_check()
    if state.distributed_type != DistributedType.XLA:
        central_dl_preparation_check()
        custom_sampler_check()
        check_seedable_sampler()
    if state.num_processes > 1: check_seedable_sampler_in_batch_sampler_shard()
    if state.distributed_type == DistributedType.DEEPSPEED: return
    training_check(use_seedable_sampler=False)
    training_check(use_seedable_sampler=True)
    test_trigger()
    if is_pytest_available(): test_reinstantiated_state()
    state.destroy_process_group()
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
class _StoreAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        new_option_strings = []
        for option_string in self.option_strings:
            new_option_strings.append(option_string)
            if "_" in option_string[2:]: new_option_strings.append(option_string.replace("_", "-"))
        self.option_strings = new_option_strings
    def __call__(self, parser, namespace, values, option_string=None): setattr(namespace, self.dest, values)
class _StoreConstAction(_StoreAction):
    def __init__(self, option_strings, dest, const, default=None, required=False, help=None): super().__init__(option_strings=option_strings, dest=dest, nargs=0, const=const, default=default, required=required, help=help)
    def __call__(self, parser, namespace, values, option_string=None): setattr(namespace, self.dest, self.const)
class _StoreTrueAction(_StoreConstAction):
    def __init__(self, option_strings, dest, default=None, required=False, help=None): super().__init__(option_strings=option_strings, dest=dest, const=True, default=default, required=required, help=help)
class CustomArgumentGroup(argparse._ArgumentGroup):
    def _add_action(self, action):
        args = vars(action)
        if isinstance(action, argparse._StoreTrueAction): action = _StoreTrueAction(args["option_strings"], args["dest"], args["default"], args["required"], args["help"])
        elif isinstance(action, argparse._StoreConstAction): action = _StoreConstAction(args["option_strings"], args["dest"], args["const"], args["default"], args["required"], args["help"])
        elif isinstance(action, argparse._StoreAction): action = _StoreAction(**args)
        action = super()._add_action(action)
        return action
class CustomArgumentParser(argparse.ArgumentParser):
    def add_argument(self, *args, **kwargs):
        if "action" in kwargs:
            if kwargs["action"] == "store_true": kwargs["action"] = _StoreTrueAction
        else: kwargs["action"] = _StoreAction
        super().add_argument(*args, **kwargs)
    def add_argument_group(self, *args, **kwargs):
        group = CustomArgumentGroup(self, *args, **kwargs)
        self._action_groups.append(group)
        return group
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import argparse
import gc
import json
import os
import torch
from datasets import load_dataset
from torch.optim import AdamW
from torch.utils.data import DataLoader
from transformers import AutoModelForSequenceClassification, AutoTokenizer, get_linear_schedule_with_warmup, set_seed
from sapiens_accelerator import Accelerator, DistributedType
from sapiens_accelerator.utils import is_mlu_available, is_musa_available, is_npu_available, is_xpu_available
from sapiens_accelerator.utils.deepspeed import DummyOptim, DummyScheduler
MAX_GPU_BATCH_SIZE = 16
EVAL_BATCH_SIZE = 32
def b2mb(x): return int(x / 2**20)
class TorchTracemalloc:
    def __enter__(self):
        gc.collect()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
            torch.cuda.reset_max_memory_allocated()
            self.begin = torch.cuda.memory_allocated()
        elif is_mlu_available():
            torch.mlu.empty_cache()
            torch.mlu.reset_max_memory_allocated()
            self.begin = torch.mlu.memory_allocated()
        elif is_musa_available():
            torch.musa.empty_cache()
            torch.musa.reset_max_memory_allocated()
            self.begin = torch.musa.memory_allocated()
        elif is_npu_available():
            torch.npu.empty_cache()
            torch.npu.reset_max_memory_allocated()
            self.begin = torch.npu.memory_allocated()
        elif is_xpu_available():
            torch.xpu.empty_cache()
            torch.xpu.reset_max_memory_allocated()
            self.begin = torch.xpu.memory_allocated()
        return self
    def __exit__(self, *exc):
        gc.collect()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
            self.end = torch.cuda.memory_allocated()
            self.peak = torch.cuda.max_memory_allocated()
        elif is_mlu_available():
            torch.mlu.empty_cache()
            torch.mlu.memory_allocated()
            self.begin = torch.mlu.max_memory_allocated()
        elif is_musa_available():
            torch.musa.empty_cache()
            torch.musa.memory_allocated()
            self.begin = torch.musa.max_memory_allocated()
        elif is_npu_available():
            torch.npu.empty_cache()
            self.end = torch.npu.memory_allocated()
            self.peak = torch.npu.max_memory_allocated()
        elif is_xpu_available():
            torch.xpu.empty_cache()
            self.end = torch.xpu.memory_allocated()
            self.peak = torch.xpu.max_memory_allocated()
        self.used = b2mb(self.end - self.begin)
        self.peaked = b2mb(self.peak - self.begin)
def get_dataloaders(accelerator: Accelerator, batch_size: int = 16, model_name: str = "bert-base-cased", n_train: int = 320, n_val: int = 160):
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    datasets = load_dataset("glue", "mrpc", split={"train": f"train[:{n_train}]", "validation": f"validation[:{n_val}]"})
    def tokenize_function(examples):
        outputs = tokenizer(examples["sentence1"], examples["sentence2"], truncation=True, max_length=None)
        return outputs
    tokenized_datasets = datasets.map(tokenize_function, batched=True, remove_columns=["idx", "sentence1", "sentence2"], load_from_cache_file=False)
    tokenized_datasets = tokenized_datasets.rename_column("label", "labels")
    def collate_fn(examples):
        if accelerator.distributed_type == DistributedType.XLA: return tokenizer.pad(examples, padding="max_length", max_length=128, return_tensors="pt")
        return tokenizer.pad(examples, padding="longest", return_tensors="pt")
    train_dataloader = DataLoader(tokenized_datasets["train"], shuffle=True, collate_fn=collate_fn, batch_size=batch_size)
    eval_dataloader = DataLoader(tokenized_datasets["validation"], shuffle=False, collate_fn=collate_fn, batch_size=EVAL_BATCH_SIZE)
    return train_dataloader, eval_dataloader
def training_function(config, args):
    accelerator = Accelerator()
    lr = config["lr"]
    num_epochs = int(config["num_epochs"])
    seed = int(config["seed"])
    batch_size = int(config["batch_size"])
    model_name = args.model_name_or_path
    set_seed(seed)
    train_dataloader, eval_dataloader = get_dataloaders(accelerator, batch_size, model_name, args.n_train, args.n_val)
    model = AutoModelForSequenceClassification.from_pretrained(model_name, return_dict=True)
    optimizer_cls = (AdamW if accelerator.state.deepspeed_plugin is None or "optimizer" not in accelerator.state.deepspeed_plugin.deepspeed_config else DummyOptim)
    optimizer = optimizer_cls(params=model.parameters(), lr=lr)
    if accelerator.state.deepspeed_plugin is not None: gradient_accumulation_steps = accelerator.state.deepspeed_plugin.deepspeed_config["gradient_accumulation_steps"]
    else: gradient_accumulation_steps = 1
    max_training_steps = (len(train_dataloader) * num_epochs) // gradient_accumulation_steps
    if (accelerator.state.deepspeed_plugin is None or "scheduler" not in accelerator.state.deepspeed_plugin.deepspeed_config): lr_scheduler = get_linear_schedule_with_warmup(optimizer=optimizer, num_warmup_steps=0, num_training_steps=max_training_steps)
    else: lr_scheduler = DummyScheduler(optimizer, total_num_steps=max_training_steps, warmup_num_steps=0)
    model, optimizer, train_dataloader, eval_dataloader, lr_scheduler = accelerator.prepare(model, optimizer, train_dataloader, eval_dataloader, lr_scheduler)
    overall_step = 0
    starting_epoch = 0
    train_total_peak_memory = {}
    for epoch in range(starting_epoch, num_epochs):
        with TorchTracemalloc() as tracemalloc:
            model.train()
            for step, batch in enumerate(train_dataloader):
                outputs = model(**batch)
                loss = outputs.loss
                loss = loss / gradient_accumulation_steps
                accelerator.backward(loss)
                if step % gradient_accumulation_steps == 0:
                    optimizer.step()
                    lr_scheduler.step()
                    optimizer.zero_grad()
                overall_step += 1
        accelerator.print(f"Memory before entering the train : {b2mb(tracemalloc.begin)}")
        accelerator.print(f"Memory consumed at the end of the train (end-begin): {tracemalloc.used}")
        accelerator.print(f"Peak Memory consumed during the train (max-begin): {tracemalloc.peaked}")
        accelerator.print(f"Total Peak Memory consumed during the train (max): {tracemalloc.peaked + b2mb(tracemalloc.begin)}")
        train_total_peak_memory[f"epoch-{epoch}"] = tracemalloc.peaked + b2mb(tracemalloc.begin)
        if args.peak_memory_upper_bound is not None: assert (train_total_peak_memory[f"epoch-{epoch}"] <= args.peak_memory_upper_bound), "Peak memory usage exceeded the upper bound"
    accelerator.wait_for_everyone()
    if accelerator.is_main_process:
        with open(os.path.join(args.output_dir, "peak_memory_utilization.json"), "w") as f: json.dump(train_total_peak_memory, f)
    accelerator.end_training()
def main():
    parser = argparse.ArgumentParser(description="Simple example of training script tracking peak GPU memory usage.")
    parser.add_argument("--model_name_or_path", type=str, default="bert-base-cased", help="", required=False)
    parser.add_argument("--output_dir", type=str, default=".", help="Optional save directory where all checkpoint folders will be stored. Default is the current working directory.")
    parser.add_argument("--peak_memory_upper_bound", type=float, default=None, help="The upper bound of peak memory usage in MB. If set, the training will throw an error if the peak memory usage exceeds this value.")
    parser.add_argument("--n_train", type=int, default=320, help="Number of training examples to use.")
    parser.add_argument("--n_val", type=int, default=160, help="Number of validation examples to use.")
    parser.add_argument("--num_epochs", type=int, default=1, help="Number of train epochs.")
    args = parser.parse_args()
    config = {"lr": 2e-5, "num_epochs": args.num_epochs, "seed": 42, "batch_size": 16}
    training_function(config, args)
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import inspect
import torch
from .state import AcceleratorState, GradientState
from .utils import DistributedType, honor_type, is_lomo_available, is_torch_xla_available
if is_torch_xla_available(): import torch_xla.core.xla_model as xm
def move_to_device(state, device):
    if isinstance(state, (list, tuple)): return honor_type(state, (move_to_device(t, device) for t in state))
    elif isinstance(state, dict): return type(state)({k: move_to_device(v, device) for k, v in state.items()})
    elif isinstance(state, torch.Tensor): return state.to(device)
    return state
class SapiensAcceleratordOptimizer(torch.optim.Optimizer):
    def __init__(self, optimizer, device_placement=True, scaler=None):
        self.optimizer = optimizer
        self.scaler = scaler
        self.accelerator_state = AcceleratorState()
        self.gradient_state = GradientState()
        self.device_placement = device_placement
        self._is_overflow = False
        if self.scaler is not None:
            self._sapiens_accelerator_step_called = False
            self._optimizer_original_step_method = self.optimizer.step
            self._optimizer_patched_step_method = patch_optimizer_step(self, self.optimizer.step)
        if device_placement:
            state_dict = self.optimizer.state_dict()
            if self.accelerator_state.distributed_type == DistributedType.XLA: xm.send_cpu_data_to_device(state_dict, self.accelerator_state.device)
            else: state_dict = move_to_device(state_dict, self.accelerator_state.device)
            self.optimizer.load_state_dict(state_dict)
    @property
    def state(self): return self.optimizer.state
    @state.setter
    def state(self, state): self.optimizer.state = state
    @property
    def param_groups(self): return self.optimizer.param_groups
    @param_groups.setter
    def param_groups(self, param_groups): self.optimizer.param_groups = param_groups
    @property
    def defaults(self): return self.optimizer.defaults
    @defaults.setter
    def defaults(self, defaults): self.optimizer.defaults = defaults
    def add_param_group(self, param_group): self.optimizer.add_param_group(param_group)
    def load_state_dict(self, state_dict):
        if self.accelerator_state.distributed_type == DistributedType.XLA and self.device_placement: xm.send_cpu_data_to_device(state_dict, self.accelerator_state.device)
        self.optimizer.load_state_dict(state_dict)
    def state_dict(self): return self.optimizer.state_dict()
    def zero_grad(self, set_to_none=None):
        if self.gradient_state.sync_gradients:
            accept_arg = "set_to_none" in inspect.signature(self.optimizer.zero_grad).parameters
            if accept_arg:
                if set_to_none is None: set_to_none = True
                self.optimizer.zero_grad(set_to_none=set_to_none)
            else:
                if set_to_none is not None: raise ValueError("`set_to_none` for Optimizer.zero_grad` is not supported by this optimizer.")
                self.optimizer.zero_grad()
    def train(self):
        if hasattr(self.optimizer, "train") and callable(self.optimizer.train): self.optimizer.train()
    def eval(self):
        if hasattr(self.optimizer, "eval") and callable(self.optimizer.eval): self.optimizer.eval()
    def step(self, closure=None):
        if is_lomo_available(): from lomo_optim import AdaLomo, Lomo
        if (not self.gradient_state.is_xla_gradients_synced and self.accelerator_state.distributed_type == DistributedType.XLA):
            gradients = xm._fetch_gradients(self.optimizer)
            xm.all_reduce("sum", gradients, scale=1.0 / xm.xrt_world_size())
            self.gradient_state.is_xla_gradients_synced = True
        if is_lomo_available():
            if isinstance(self.optimizer, (Lomo, AdaLomo)): return
        if self.gradient_state.sync_gradients:
            if self.scaler is not None:
                self.optimizer.step = self._optimizer_patched_step_method
                self.scaler.step(self.optimizer, closure)
                self.scaler.update()
                if not self._sapiens_accelerator_step_called: self._is_overflow = True
                else: self._is_overflow = False
                self.optimizer.step = self._optimizer_original_step_method
                self._sapiens_accelerator_step_called = False
            else: self.optimizer.step(closure)
        if self.accelerator_state.distributed_type == DistributedType.XLA: self.gradient_state.is_xla_gradients_synced = False
    def _switch_parameters(self, parameters_map):
        for param_group in self.optimizer.param_groups: param_group["params"] = [parameters_map.get(p, p) for p in param_group["params"]]
    @property
    def step_was_skipped(self): return self._is_overflow
    def __getstate__(self):
        _ignored_keys = ["_sapiens_accelerator_step_called", "_optimizer_original_step_method", "_optimizer_patched_step_method"]
        return {k: v for k, v in self.__dict__.items() if k not in _ignored_keys}
    def __setstate__(self, state):
        self.__dict__.update(state)
        if self.scaler is not None:
            self._sapiens_accelerator_step_called = False
            self._optimizer_original_step_method = self.optimizer.step
            self._optimizer_patched_step_method = patch_optimizer_step(self, self.optimizer.step)
def patch_optimizer_step(sapiens_acceleratord_optimizer: SapiensAcceleratordOptimizer, method):
    def patched_step(*args, **kwargs):
        sapiens_acceleratord_optimizer._sapiens_accelerator_step_called = True
        return method(*args, **kwargs)
    return patched_step
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import json
import os
from collections.abc import Mapping
from typing import Dict, List, Optional, Union
import numpy as np
import torch
from safetensors import safe_open
def offload_weight(weight, weight_name, offload_folder, index=None):
    dtype = None
    if str(weight.dtype) == "torch.bfloat16":
        weight = weight.view(torch.int16)
        dtype = "bfloat16"
    array = weight.cpu().numpy()
    tensor_file = os.path.join(offload_folder, f"{weight_name}.dat")
    if index is not None:
        if dtype is None: dtype = str(array.dtype)
        index[weight_name] = {"dtype": dtype, "shape": list(array.shape)}
    if array.ndim == 0: array = array[None]
    file_array = np.memmap(tensor_file, dtype=array.dtype, mode="w+", shape=array.shape)
    file_array[:] = array[:]
    file_array.flush()
    return index
def load_offloaded_weight(weight_file, weight_info):
    shape = tuple(weight_info["shape"])
    if shape == (): shape = (1,)
    dtype = weight_info["dtype"]
    if dtype == "bfloat16": dtype = "int16"
    weight = np.memmap(weight_file, dtype=dtype, shape=shape, mode="r")
    if len(weight_info["shape"]) == 0: weight = weight[0]
    weight = torch.tensor(weight)
    if weight_info["dtype"] == "bfloat16": weight = weight.view(torch.bfloat16)
    return weight
def save_offload_index(index, offload_folder):
    if index is None or len(index) == 0: return
    offload_index_file = os.path.join(offload_folder, "index.json")
    if os.path.isfile(offload_index_file):
        with open(offload_index_file, encoding="utf-8") as f: current_index = json.load(f)
    else: current_index = {}
    current_index.update(index)
    with open(offload_index_file, "w", encoding="utf-8") as f: json.dump(current_index, f, indent=2)
def offload_state_dict(save_dir: Union[str, os.PathLike], state_dict: Dict[str, torch.Tensor]):
    os.makedirs(save_dir, exist_ok=True)
    index = {}
    for name, parameter in state_dict.items(): index = offload_weight(parameter, name, save_dir, index=index)
    save_offload_index(index, save_dir)
class PrefixedDataset(Mapping):
    def __init__(self, dataset: Mapping, prefix: str):
        self.dataset = dataset
        self.prefix = prefix
    def __getitem__(self, key): return self.dataset[f"{self.prefix}{key}"]
    def __iter__(self): return iter([key for key in self.dataset if key.startswith(self.prefix)])
    def __len__(self): return len(self.dataset)
class OffloadedWeightsLoader(Mapping):
    def __init__(self, state_dict: Dict[str, torch.Tensor] = None, save_folder: Optional[Union[str, os.PathLike]] = None, index: Mapping = None, device=None):
        if state_dict is None and save_folder is None and index is None: raise ValueError("Need either a `state_dict`, a `save_folder` or an `index` containing offloaded weights.")
        self.state_dict = {} if state_dict is None else state_dict
        self.save_folder = save_folder
        if index is None and save_folder is not None:
            with open(os.path.join(save_folder, "index.json")) as f: index = json.load(f)
        self.index = {} if index is None else index
        self.all_keys = list(self.state_dict.keys())
        self.all_keys.extend([key for key in self.index if key not in self.all_keys])
        self.device = device
    def __getitem__(self, key: str):
        if key in self.state_dict: return self.state_dict[key]
        weight_info = self.index[key]
        if weight_info.get("safetensors_file") is not None:
            device = "cpu" if self.device is None else self.device
            tensor = None
            try:
                with safe_open(weight_info["safetensors_file"], framework="pt", device=device) as f: tensor = f.get_tensor(weight_info.get("weight_name", key))
            except TypeError:
                with safe_open(weight_info["safetensors_file"], framework="pt", device="cpu") as f: tensor = f.get_tensor(weight_info.get("weight_name", key))
            if "dtype" in weight_info: tensor = tensor.to(getattr(torch, weight_info["dtype"]))
            if tensor.device != torch.device(device): tensor = tensor.to(device)
            return tensor
        weight_file = os.path.join(self.save_folder, f"{key}.dat")
        return load_offloaded_weight(weight_file, weight_info)
    def __iter__(self): return iter(self.all_keys)
    def __len__(self): return len(self.all_keys)
def extract_submodules_state_dict(state_dict: Dict[str, torch.Tensor], submodule_names: List[str]):
    result = {}
    for module_name in submodule_names: result.update({key: param for key, param in state_dict.items() if key == module_name or key.startswith(module_name + ".")})
    return result
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
	########################################################################################################################################################
	# This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
	# or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
	########################################################################################################################################################
"""
from .testing import (DEFAULT_LAUNCH_COMMAND, are_the_same_tensors, assert_exception, capture_call_output, device_count, execute_subprocess_async,
get_launch_command, memory_allocated_func, path_in_sapiens_accelerator_package, require_sapiens, require_cpu, require_cuda, require_huggingface_suite, require_mlu,
require_mps, require_multi_device, require_multi_gpu, require_multi_xpu, require_musa, require_non_cpu, require_non_torch_xla, require_non_xpu, require_npu,
require_pippy, require_single_device, require_single_gpu, require_single_xpu, require_torch_min_version, require_torchvision, require_tpu, require_transformer_engine,
require_xpu, skip, slow, torch_device)
from .training import RegressionDataset, RegressionModel, RegressionModel4XPU
from .scripts import test_script, test_sync, test_ops
"""
	########################################################################################################################################################
	# This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
	# or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
	########################################################################################################################################################
"""
